# Author:    Lisandro Dalcin
# Contact:   dalcinl@users.sourceforge.net
# Revision:  $Revision: 1.27 $
# Date:      $Date: 2006/08/09 00:35:20 $
# Copyright: This module has been placed in the public domain.

"""
Message Passing Interface module.

This module provides MPI suport to run Python scripts in parallel. It
is constructed on top of the MPI-1 specification, but provides an
object oriented interface which closely follows stantard MPI-2 C++
bindings.

Any *picklable* Python object can be communicated. There is support
for point-to-point (sends, receives) and collective (broadcasts,
scatters, gathers) communications as well as group and communicator
(inter, intra and topologies) management.

.. note:: It is recommendable to use this module with a parallelized
   version of the Python interpreter.
"""

__docformat__ = 'reStructuredText'


# --------------------------------------------------------------------
# Metadata
# --------------------------------------------------------------------

__author__    = 'Lisandro Dalcin'
__credits__   = 'MPI Forum, MPICH Team, LAM Team, Open MPI Team.'
__date__      = '02 Aug 2006'
__version__   = '0.3.1'
__revision__  = '$Id: MPI.py,v 1.27 2006/08/09 00:35:20 dalcinl Exp $'


# --------------------------------------------------------------------
# Imports
# --------------------------------------------------------------------

import mpi4py._mpi    as _mpi     # extension module
import mpi4py._op     as _op      # reduction operations
import mpi4py._pickle as _pickle  # pickling support


# --------------------------------------------------------------------
# Constants
# --------------------------------------------------------------------

UNDEFINED = _mpi.UNDEFINED
"""Undefined integer value, `MPI_UNDEFINED`."""

PROC_NULL  = _mpi.PROC_NULL  
"""Special process rank for `MPI_SEND` and `MPI_RECV`,
`MPI_PROC_NULL`."""

ANY_SOURCE = _mpi.ANY_SOURCE
"""Wildcard source value for `MPI_RECV`, `MPI_ANY_SOURCE`."""

ANY_TAG    = _mpi.ANY_TAG
"""Wildcard tag value for `MPI_RECV`, `MPI_ANY_TAG`."""

TAG_DEFAULT   = 0               # normal send/recv
TAG_STREAM    = _mpi.TAG_UB/2-1 # ports with streams
TAG_INTERCOMM = _mpi.TAG_UB/2-2 # Intercomm creation
"""Some tunable tags used for this module.

- ``TAG_DEFAULT``:   default tag used in normal sends and receives.
- ``TAG_STREAM``:    default tag used in ports with *streams*.
- ``TAG_INTERCOMM``: default tag used in intercommunicator creation.
"""

IDENT     =  _mpi.IDENT      # same handles
CONGRUENT =  _mpi.CONGRUENT  # same groups and rank orderings
SIMILAR   =  _mpi.SIMILAR    # same groups
UNEQUAL   =  _mpi.UNEQUAL    # different groups
"""Result of group and communicator comparisons,
`MPI_IDENT`, `MPI_CONGRUENT`, `MPI_SIMILAR` and `MPI_UNEQUAL`."""



# --------------------------------------------------------------------
# Datatype
# --------------------------------------------------------------------

class Datatype(_mpi.Datatype):

    """
    Datatype class.
    """
    
    def __new__(cls, datatype=None):
        return _mpi.Datatype.__new__(cls, datatype)

    def __init__(self, datatype=None):
        _mpi.Datatype.__init__(self, datatype)

    # Datatype Accessors
    # ------------------
    
    def Get_extent(self):
        """
        Get lower bound and extent of datatype as a tuple (lb,extent).
        """
        return _mpi.type_get_extent(self)

    def Get_size(self):
        """
        Returns the number of bytes occupied by entries in the
        datatype.
        """
        return _mpi.type_size(self)

    # Datatype Constructors
    # ---------------------
    
    def Dup(self):
        """
        Duplicates a datatype.
        """
        newdt = _mpi.type_contiguous(1, self)
        return Datatype(newdt)

    def Create_contiguous(self, count):
        """
        Creates a contiguous datatype.
        """
        newdt = _mpi.type_contiguous(count, self)
        return Datatype(newdt)

    def Create_vector(self, count, blocklength, stride):
        """
        Creates a vector (strided) datatype.
        """
        newdt = _mpi.type_vector(count, blocklength, stride, self)
        return Datatype(newdt)
   
    # Use of Derived Datatypes
    # ------------------------
    
    def Commit(self):
        """
        Commits the datatype.
        """
        _mpi.type_commit(self)

    def Free(self):
        """
        Frees the datatype.
        """
        _mpi.type_free(self)

    # Properties
    # ----------
    
    extent = property(_mpi.type_extent, doc="Datatype extent.")
    lb     = property(_mpi.type_lb,     doc="Datatype lower bound.")
    ub     = property(_mpi.type_ub,     doc="Datatype upper bound.")
    
    size   = property(Get_size)

# Predefined datatype handles

DATATYPE_NULL  = Datatype(_mpi.DATATYPE_NULL) 
"""Null datatype handle, `MPI_DATATYPE_NULL`"""

CHAR            = Datatype(_mpi.CHAR)
SHORT           = Datatype(_mpi.SHORT)
INT             = Datatype(_mpi.INT)
LONG            = Datatype(_mpi.LONG)
UNSIGNED_CHAR   = Datatype(_mpi.UNSIGNED_CHAR)
UNSIGNED_SHORT  = Datatype(_mpi.UNSIGNED_SHORT)
UNSIGNED        = Datatype(_mpi.UNSIGNED)
UNSIGNED_LONG   = Datatype(_mpi.UNSIGNED_LONG)
FLOAT           = Datatype(_mpi.FLOAT)
DOUBLE          = Datatype(_mpi.DOUBLE)
LONG_DOUBLE     = Datatype(_mpi.LONG_DOUBLE)
BYTE            = Datatype(_mpi.BYTE)
PACKED          = Datatype(_mpi.PACKED)
"""Elementary datatypes."""

FLOAT_INT       = Datatype(_mpi.FLOAT_INT)
DOUBLE_INT      = Datatype(_mpi.DOUBLE_INT)
LONG_INT        = Datatype(_mpi.LONG_INT)
TWOINT          = Datatype(_mpi.TWOINT)
SHORT_INT       = Datatype(_mpi.SHORT_INT)
LONG_DOUBLE_INT = Datatype(_mpi.LONG_DOUBLE_INT)
"""Datatypes for reduction functions."""
 
UB              = Datatype(_mpi.UB)
LB              = Datatype(_mpi.LB)
"""Special datatypes for constructing derived datatypes."""




# --------------------------------------------------------------------
# Status
# --------------------------------------------------------------------

class Status(_mpi.Status):

    """
    Status class.
    """
    
    def __new__(cls, status=None):
        return _mpi.Status.__new__(cls, status)

    def __init__(self, status=None):
        _mpi.Status.__init__(self, status)

    def Get_source(self):
        """
        Gets message source.
        """
        return self.MPI_SOURCE
    
    def Get_tag(self):
        """
        Gets message tag.
        """
        return self.MPI_TAG
    
    def Get_error(self):
        """
        Gets message error.
        """
        return self.MPI_ERROR

    def Set_source(self, source):
        """
        Sets message surce.
        """
        self.MPI_SOURCE = source
    
    def Set_tag(self, tag):
        """
        Sets message tag.
        """
        self.MPI_TAG = tag
    
    def Set_error(self, error):
        """
        Sets message error.
        """
        self.MPI_ERROR = error
    
    def Get_count(self, datatype=None):
        """
        Gets the number of *top level* elements.

        .. note:: Currently, `datatype` argument is meaningless,
           default value is `MPI_BYTE`.
        """
        if datatype is None:
            datatype = _mpi.BYTE
        return _mpi.get_count(self, datatype)

    def Get_elements(self, datatype=None):
        """
        Gets the number of basic elements in a datatype.

        .. note:: Currently, `datatype` argument is meaningless,
           default value is `MPI_BYTE`.
        """
        if datatype is None:
            datatype = _mpi.BYTE
        return _mpi.get_elements(self, datatype)
    
    def Is_cancelled(self):
        """
        Tests to see if a request was cancelled.
        """
        return _mpi.test_cancelled(self)
    
    # Properties
    # ----------

    source = property(Get_source, Set_source, doc='message source')
    tag    = property(Get_tag,    Set_tag,    doc='message tag')
    error  = property(Get_error,  Set_error,  doc='message error')

   
STATUS_DEFAULT = Status()
"""Default status handle used in sends and receives."""
    



# --------------------------------------------------------------------
# Request
# --------------------------------------------------------------------

class Request(_mpi.Request):

    """
    Request class.
    """
    
    def __new__(cls, request=None):
        return _mpi.Request.__new__(cls, request)

    def __init__(self, request=None):
        _mpi.Request.__init__(self, request)
    
    def Wait(self, status=None):
        """
        Waits for an MPI send or receive to complete.
        """
        if status is None:
            status = STATUS_DEFAULT
        return _mpi.wait(self, status)
    
    def Test(self, status=None):
        """
        Tests for the completion of a send or receive.
        """
        if status is None:
            status = STATUS_DEFAULT
        return _mpi.test(self, status)
        
    def Free(self):
        """
        Frees a communication request object.
        """
        return _mpi.request_free(self)
        
    def Cancel(self):
        """
        Cancels a communication request.
        """
        return _mpi.cancel(self)


REQUEST_NULL = Request(_mpi.REQUEST_NULL)
"""Null request handle, `MPI_REQUEST_NULL`."""




# --------------------------------------------------------------------
# Operations
# --------------------------------------------------------------------

class Op(_mpi.Op):

    """
    Op class.
    """

    def __new__(cls, op=None):
        newop = _mpi.Op.__new__(cls, op)
        if newop is not op:
            def nullop(*args): _mpi._raise(_mpi.ERR_OP)
            newop.__function = nullop
            newop.__commute  = False
        return newop

    def __init__(self, op=None):
        _mpi.Op.__init__(self, op)

    def __call__(self, x, y):
        """
        Calls the user-defined combination function handle.
        """
        return self.__function(x, y)

    def Init(self, function, commute=False):
        """
        Creates a user-defined combination function handle.
        """
        if not callable(function):
            raise TypeError('first argument must be a callable object')
        if type(commute) is not bool:
            raise TypeError('second argument must be boolean')
        self.__function = function
        self.__commute  = commute
        
    def Free(self):
        """
        Frees a user-defined combination function handle.
        """
        Op.__init__(self)
            

OP_NULL = Op(_mpi.OP_NULL)
"""Null operation handle, `MPI_OP_NULL`."""

MAX    = Op(_mpi.MAX)
MIN    = Op(_mpi.MIN)
SUM    = Op(_mpi.SUM)
PROD   = Op(_mpi.PROD)
LAND   = Op(_mpi.LAND)
BAND   = Op(_mpi.BAND)
LOR    = Op(_mpi.LOR)
BOR    = Op(_mpi.BOR)
LXOR   = Op(_mpi.LXOR)
BXOR   = Op(_mpi.BXOR)
MAXLOC = Op(_mpi.MAXLOC)
MINLOC = Op(_mpi.MINLOC)
"""Predefined operations."""

MAX.Init(_op.MAX, True)
MIN.Init(_op.MIN, True)
SUM.Init(_op.SUM, True)
PROD.Init(_op.PROD, True)
LAND.Init(_op.LAND, True)
BAND.Init(_op.BAND, True)
LOR.Init(_op.LOR, True)
BOR.Init(_op.BOR, True)
LXOR.Init(_op.LXOR, True)
BXOR.Init(_op.BXOR, True)
MAXLOC.Init(_op.MAXLOC, True)
MINLOC.Init(_op.MINLOC, True)


# --------------------------------------------------------------------
# Group
# --------------------------------------------------------------------

class Group(_mpi.Group):

    """
    Group class.
    """

    def __new__(cls, group=None):
        return _mpi.Group.__new__(cls, group)

    def __init__(self, group=None):
        _mpi.Group.__init__(self, group)

    # Group Accessors
    # ---------------
    
    def Get_size(self):
        """
        Returns the size of a group.
        """
        return _mpi.group_size(self)

    def Get_rank(self):
        """
        Returns the rank of this process in the group.
        """
        return _mpi.group_rank(self)

    ## @staticmethod
    def Translate_ranks(group1, ranks1, group2):
        """
        Translates the ranks of processes in one group to those in
        another group.
        """
        return _mpi.group_transl_rank(group1, ranks1, group2)

    Translate_ranks = staticmethod(Translate_ranks)

    ## @staticmethod
    def Compare(group1, group2):
        """
        Compare two groups.
        """
        return _mpi.group_compare(group1, group2)

    Compare = staticmethod(Compare)

    # Group Constructors
    # ------------------
    
    ## @staticmethod
    def Union(group1, group2):
        """
        Produces a group by combining two groups.
        """
        newgroup = _mpi.group_union(group1, group2)
        return Group(newgroup)

    Union = staticmethod(Union)

    ## @staticmethod
    def Intersect(group1, group2):
        """
        Produces a group as the intersection of two existing groups.
        """
        newgroup = _mpi.group_intersection(group1, group2)
        return Group(newgroup)        

    Intersect = staticmethod(Intersect)
    
    ## @staticmethod
    def Difference(group1, group2):
        """
        Makes a group from the difference of two groups.
        """
        newgroup = _mpi.group_difference(group1, group2)
        return Group(newgroup)

    Difference = staticmethod(Difference)

    def Incl(self, ranks):
        """
        Produces a group by reordering an existing group and taking
        only listed members.
        """
        newgroup = _mpi.group_incl(self, ranks)
        return Group(newgroup)

    def Excl(self, ranks):
        """
        Produces a group by reordering an existing group and taking
        only unlisted members.
        """
        newgroup = _mpi.group_excl(self, ranks)
        return Group(newgroup)
    
    def Range_incl(self, ranks):
        """
        Creates a new group from ranges of ranks in an existing group.
        """
        newgroup = _mpi.group_range_incl(self, ranks)
        return Group(newgroup)
    
    def Range_excl(self, ranks):
        """
        Produces a group by excluding ranges of pro- cesses from an
        existing group.
        """
        newgroup = _mpi.group_range_excl(self, ranks)
        return Group(newgroup)    

    # Group Destructor
    # ----------------
    
    def Free(self):
        """
        Frees a group.
        
        .. note:: `self` will have a handle to `MPI_GROUP_NULL` upon
           successful return.
        """
        return _mpi.group_free(self)

    # Properties
    # ----------
    
    size = property(Get_size, doc=None)
    rank = property(Get_rank, doc=None)


GROUP_NULL = Group(_mpi.GROUP_NULL)
"""Null group handle, `MPI_GROUP_NULL`."""

GROUP_EMPTY = Group(_mpi.GROUP_EMPTY)
"""Empty group handle, `MPI_GROUP_EMPTY`."""




# --------------------------------------------------------------------
# Communicator
# --------------------------------------------------------------------

class Comm(_mpi.Comm):

    """
    Communicator class.
    """

    def __new__(cls, comm=None):
        return _mpi.Comm.__new__(cls, comm)

    def __init__(self, comm=None):
        _mpi.Comm.__init__(self, comm)
        ## if self==_mpi.COMM_NULL: return
        ## if type(self) is Comm:
        ##    raise TypeError, 'virtual base class'


    # Group
    # -----
    
    def Get_group(self):
        """
        Accesses the group associated with a communicator.
        """
        group = _mpi.comm_group(self)
        return Group(group)

    # Tests
    # -----

    def Is_inter(self):
        """
        Tests to see if a comm is an intercommunicator.
        """
        return _mpi.comm_test_inter(self)

    def Is_intra(self):
        """
        Tests to see if a comm is an intracommunicator.
        """
        return not _mpi.comm_test_inter(self)
    
    def Get_topology(self):
        """
        Determines the type of topology (if any) associated with a
        communicator.
        """
        return _mpi.topo_test(self)

    # Communicator Accessors
    # ----------------------
    
    def Get_size(self):
        """
        Determines the size of the group associated with a
        communicator.
        """
        return _mpi.comm_size(self)

    def Get_rank(self):
        """
        Determines the rank of the calling process in the
        communicator.
        """
        return _mpi.comm_rank(self)

    ## @staticmethod
    def Compare(comm1, comm2):
        """
        Compare two communicators.
        """
        return _mpi.comm_compare(comm1, comm2)

    Compare = staticmethod(Compare)

    # Communicator Constructors
    # -------------------------
    
    def Clone(self):
        """
        Duplicate an existing communicator with all its cached
        information.
        """
        newcomm = _mpi.comm_dup(self)
        return self.__class__(newcomm)

    # Communicator Destructors
    # ------------------------
    
    def Free(self):
        """
        Marks the communicator object for deallocation.

        .. note:: `self` will have a handle to `MPI_COMM_NULL` upon
           successful return.
        """
        _mpi.comm_free(self)
        
    # Point to Point communication
    # ----------------------------

    def Send(self, buff, dest=0, tag=None):
        """
        Performs a basic send.

        .. note:: This function may block until the message is
           received. Whether or not `MPI_SEND` blocks depends on
           several factors and is implementation dependent.
        """
        if tag is None:
            tag = TAG_DEFAULT
        buff = _pickle.dump(buff)
        _mpi.send_string(buff, dest, tag, self)
        return None
    
    
    def Recv(self, source=0, tag=None, status=None):
        """
        Basic receive.

        .. note:: This function blocks until the message is received.
        """
        if tag    is None:
            tag    = TAG_DEFAULT
        if status is None:
            status = STATUS_DEFAULT
        buff = _mpi.recv_string(source, tag, self, status)
        buff = _pickle.load(buff)
        return buff
    
    def Sendrecv(self, buff, dest=0, source=0,
                 sendtag=None, recvtag=None,
                 status=None):
        """
        Sends and receives a message.

        .. note:: This function is guaranteed not to deadlock in
	   situations where pairs of blocking sends and receives may
	   deadlock.

        .. caution:: A common mistake when using this function is to
	   mismatch the tags with the source and destination ranks,
	   which can result in deadlock.
        """
        if sendtag is None:
            sendtag = TAG_DEFAULT
        if recvtag is None:
            recvtag = TAG_DEFAULT
        if status  is None:
            status  = STATUS_DEFAULT
        buff = _pickle.dump(buff)
        buff = _mpi.sendrecv_string(buff,
                                    dest, source,
                                    sendtag, recvtag,
                                    self, status)
        buff = _pickle.load(buff)
        return buff

    def Probe(self, source, tag=None, status=None):
        """
        Blocking test for a message.
        """
        if tag is None:
            tag    = TAG_DEFAULT
        if status is None:
            status = STATUS_DEFAULT
        _mpi.comm_probe(source, tag, self, status)

    # Error handling
    # --------------
    
    def Get_errhandler(self):
        """
        Gets the error handler for a communicator.
        """
        eh = _mpi.comm_get_errhandler(self)
        return Errhandler(eh)

    def Set_errhandler(self, errhandler):
        """
        Sets the error handler for a communicator.
        """
        _mpi.comm_set_errhandler(self, errhandler)

    # 
    
    def Abort(self, errorcode=0):
        """
        Terminates MPI execution environment.

        .. warning:: This is a direct call, use it with care!!!.
        """
        _mpi.comm_abort(self, errorcode)

    # Port (OOMPI-like)
    # ----

    class Port(object):
        """Port (Communicator nested class)."""
        def __init__(self, comm, pid):
            if pid not in xrange(len(comm)):
                raise ValueError('port number out of range')
            self.__comm = comm
            self.__pid  = pid
        def Send(self, buff, tag=None):
            """Send to port."""
            if tag is None: tag = TAG_DEFAULT
            return self.comm.Send(buff, self.pid, tag)
        def Recv(self, tag=None):
            """Receive from port."""
            if tag is None: tag = TAG_DEFAULT
            return self.comm.Recv(self.pid, tag)
        def __lshift__(self, stream):
            """Send sequence to port."""
            if type(stream) is not list:
                raise TypeError('input stream must be a list')
            self.Send(stream, TAG_STREAM)
        def __rshift__(self, stream):
            """Receive sequence from port."""
            if type(stream) is not list:
                raise TypeError('output stream must be a list')
            stream.extend(self.Recv(TAG_STREAM))
        def comm(self): return self.__comm
        def pid(self):  return self.__pid
        comm = property(comm, doc='associated communicator')
        pid  = property(pid,  doc='associated process *id*')
        
    # Sequence Methods
    # ----------------

    def __len__(self):
        """
        Determines the number of ports of a communicator.
        """
        if self == _mpi.COMM_NULL:
            return 0
        else:
            return self.Get_size()

    def __getitem__(self, i):
        """
        Access the i'th port of a communicator.
        """
        if type(i) is not int:
            raise TypeError('indices must be integers')
        if i < 0:
            i += len(self)
        if i < 0 or i >= len(self):
            raise IndexError('index out of range')
        return self.__class__.Port(self, i)

    def __iter__(self):
        """
        Returns an iterator.
        """
        i = 0
        while i < len(self):
            yield self.__class__.Port(self, i)
            i += 1

    # Properties
    # ----------

    size = property(Get_size,     doc=None)
    rank = property(Get_rank,     doc=None)

    is_intra  = property(Is_intra,     doc=None)
    is_inter  = property(Is_inter,     doc=None)
    topology  = property(Get_topology, doc=None)

    errhandler = property(Get_errhandler, Set_errhandler, doc=None)


COMM_NULL = Comm(_mpi.COMM_NULL)
"""Null communicator handle, `MPI_COMM_NULL`."""




# --------------------------------------------------------------------
# Intracommunicator
# --------------------------------------------------------------------

class Intracomm(Comm):

    """
    Intracommunicator class.
    """

    def __new__(cls, comm=None):
        return Comm.__new__(cls, comm)

    def __init__(self, comm=None):
        Comm.__init__(self, comm)
        if self != _mpi.COMM_NULL:
            if self.Is_inter():
                raise TypeError('intercommunicator is not allowed')
    
    # Communicator Constructors
    # -------------------------

    def Dup(self):
        """
        Duplicate intracommunicator.
        """
        newcomm = _mpi.comm_dup(self)
        return Intracomm(newcomm)
    
    def Create(self, group):
        """
        Create intracommunicator from group.
        """
        newcomm = _mpi.comm_create(self, group)
        return Intracomm(newcomm)
    
    def Split(self, color, key=0):
        """
        Split communicator by color and key.
        """
        newcomm = _mpi.comm_split(self, color, key)
        return Intracomm(newcomm)

    def Create_intercomm(self,      local_leader,
                         peer_comm, remote_leader,
                         tag=None):
        """
        Create intercommunicator.
        """
        if tag is None:
            tag = TAG_INTERCOMM
        newcomm = _mpi.intercomm_create(self,      local_leader,
                                        peer_comm, remote_leader,
                                        tag)
        return Intercomm(newcomm)

    def Create_cart(self, dims, periods=None, reorder=False):
        """
        Create cartesian communicator.
        """
        if periods is None:
            periods = [False] * len(dims)
        newcomm = _mpi.cart_create(self,
                                   dims, periods,
                                   reorder)
        return Cartcomm(newcomm)
    
    def Create_graph(self, index, edges, reorder=False):
        """
        Create graph communicator.
        """
        newcomm = _mpi.graph_create(self,
                                    index, edges,
                                    reorder)
        return Graphcomm(newcomm)

    
    # Barrier Synchronization
    # -----------------------

    def Barrier(self):
        """
        Barrier synchronization.
        """
        _mpi.barrier(self)
        
    # Global Communication Functions
    # ------------------------------

    def Bcast(self, buff=None, root=0):
        """
        Broadcasts a message from one process to all other processes
        in a group.
        """
        if self.Get_rank() == root:
            buff = _pickle.dump(buff)
        else:
            buff = ''
        buff = _mpi.bcast_string(buff, root, self)
        buff = _pickle.load(buff)
        return buff

    def Gather(self, buff, root=0):
        """
        Gathers together values from a group of processes.
        """
        buff = _pickle.dump(buff)
        buff = _mpi.gather_string(buff, root, self)
        if self.Get_rank() == root:
            buff = [_pickle.load(b) for b in buff]
        else:
            buff = None
        return buff

    def Scatter(self, buff=None, root=0):
        """
        Scatters data from one process to all other processes in a
        group.
        """
        if self.Get_rank() == root:
            buff = [_pickle.dump(b) for b in buff]
        else:
            buff = ''
        buff = _mpi.scatter_string(buff, root, self)
        buff = _pickle.load(buff)
        return buff
            

    def Allgather(self, buff):
        """
        Gather to All, gathers data from all processes and distribute
        it to all other processes in a group.
        """
        buff = _pickle.dump(buff)
        buff = _mpi.allgather_string(buff, self)
        buff = [_pickle.load(b) for b in buff]
        return buff


    def Alltoall(self, buff):
        """
        All to All Scatter/Gather, sends data from all to all
        processes in a group.
        """
        buff = [_pickle.dump(b) for b in buff]
        buff = _mpi.alltoall_string(buff, self)
        buff = [_pickle.load(b) for b in buff]
        return buff


    # Global Reduction Operations (naive implementations)
    # ---------------------------
    
    def Reduce(self, buff, root=0, op=SUM):
        """
        Reduce.
        """
        if not callable(op):
            raise ValueError('operation is not callable')
        buff = self.Gather(buff, root)
        if self.Get_rank() == root:
            if op in [MAXLOC, MINLOC]:
                buff = zip(buff, xrange(self.Get_size()))
            buff = reduce(op, buff)
        return buff

    def Allreduce(self, buff, op=SUM):
        """
        All Reduce.
        """
        if not callable(op):
            raise ValueError('operation is not callable')
        buff = self.Reduce(buff, 0, op)
        buff = self.Bcast(buff, 0)
        return buff
    
    ## def Reduce_scatter(self, buff, op=SUM):
    ##     """
    ##     Reduce-Scatter. - Not implemented
    ##     """
    ##     raise NotImplementedError

    def Scan(self, buff, op=SUM):
        """
        Inclusive Scan.
        """
        if not callable(op):
            raise ValueError('operation is not callable')
        buff = self.Gather(buff, 0)
        if self.Get_rank() == 0:
            if op in [MAXLOC, MINLOC]:
                buff = zip(buff, xrange(self.Get_size()))
            for i in xrange(1, self.Get_size()):
                buff[i] = op(buff[i-1], buff[i])
        return self.Scatter(buff, 0)
        
    # Port
    # ----

    class Port(Comm.Port):
        """Port (Intracommunicator nested class)."""
        def Bcast(self, buff=None):
            """Broadcast from port."""
            return self.comm.Bcast(buff, self.pid)
        def Gather(self, buff):
            """Gather to port."""
            return self.comm.Gather(buff, self.pid)
        def Scatter(self, buff=None):
            """Scatter from port."""
            return self.comm.Scatter(buff, self.pid)
        def Reduce(self, buff, op):
            """Reduce to port."""
            return self.comm.Reduce(buff, self.pid, op)


__COMM_WORLD__ = Intracomm(_mpi.COMM_WORLD)
__COMM_SELF__  = Intracomm(_mpi.COMM_SELF)
"""Predefined communicators. Original instances."""

COMM_WORLD = WORLD =  __COMM_WORLD__.Dup()
COMM_SELF  = SELF  =  __COMM_SELF__.Dup()
"""Predefined communicators. Duplicated instances (for safety)."""



# --------------------------------------------------------------------
# Virtual Topologies
# --------------------------------------------------------------------

# Types

CART  = _mpi.CART
"""Cartesian topology, `MPI_CART`."""

GRAPH = _mpi.GRAPH
"""Graph topology, `MPI_GRAPH`."""


# --------------------------------------------------------------------
# Cartesian
# --------------------------------------------------------------------

# Cartesian Convenience Function

def Compute_dims(nnodes, dims):
    """
    Returns a balanced distribution of processes per coordinate
    direction.
    """
    return _mpi.dims_create(nnodes, dims)

class Cartcomm(Intracomm):

    """
    Cartesian topology class.
    """

    def __new__(cls, comm=None):
        return Intracomm.__new__(cls, comm)

    def __init__(self, comm=None):
        Intracomm.__init__(self, comm)
        if self != _mpi.COMM_NULL:
            if self.Get_topology() != _mpi.CART:
                raise TypeError('non-cartesian communicator '
                                'is not allowed')

    # Communicator Constructors
    # -------------------------

    def Dup(self):
        """
        Duplicate cartesian communicator.
        """
        newcomm = _mpi.comm_dup(self)
        return Cartcomm(newcomm)

    # Cartesian Inquiry Functions
    # ---------------------------
    
    def Get_dim(self):
        """
        Returns number of dimensions.
        """
        return _mpi.cartdim_get(self)
    
    def Get_topo(self):
        """
        Returns information on the cartesian topology.
        """
        return _mpi.cart_get(self)

    # Cartesian Translator Functions
    # ------------------------------
    
    def Get_cart_rank(self, coords):
        """
        Translate logical coordinates to ranks.
        """
        return _mpi.cart_rank(self, coords)

    def Get_coords(self, rank):
        """
        Translate ranks to logical coordinates.
        """
        return _mpi.cart_coords(self, rank)

    # Cartesian Shift Function
    # ------------------------

    def Shift(self, direction, disp):
        """
        Returns a tuple (source,dest) of process ranks for data
        shifting with Comm.Sendrecv().
        """
        return _mpi.cart_shift(self, direction, disp)

    # Cartesian Partition Function
    # ----------------------------

    def Sub(self, remain_dims):
        """
        Returns cartesian communicators that form lower-dimensional
        subgrids.
        """
        newcomm = _mpi.cart_sub(self, remain_dims)
        return Cartcomm(newcomm)

    # Cartesian Low-Level Functions
    # -----------------------------

    def Map(self, dims, periods):
        """
        Returns an optimal placement for the calling process on the
        physical machine.
        """
        return _mpi.cart_map(self, dims, periods)

    # Properties
    # ----------

    dim  = property(Get_dim,  doc=None)
    topo = property(Get_topo, doc=None)



# --------------------------------------------------------------------
# Graph 
# --------------------------------------------------------------------

class Graphcomm(Intracomm):

    """
    Graph topology class.
    """

    def __new__(cls, comm=None):
        return Intracomm.__new__(cls, comm)

    def __init__(self, comm=None):
        Intracomm.__init__(self, comm)
        if self != _mpi.COMM_NULL:
            if self.Get_topology() != _mpi.GRAPH:
                raise TypeError('non-graph communicator '
                                'is not allowed')

    # Communicator Constructors
    # -------------------------

    def Dup(self):
        """
        Duplicate graph communicator.
        """
        newcomm = _mpi.comm_dup(self)
        return Graphcomm(newcomm)

    # Graph Inquiry Functions
    # -----------------------
    
    def Get_dims(self):
        """
        Returns the number of nodes and edges.
        """
        return _mpi.graphdims_get(self)

    def Get_topo(self):
        """
        Returns index and edges.
        """
        return _mpi.graph_get(self)

    # Graph Information Functions
    # ---------------------------
    
    def Get_neighbors_count(self, rank):
        """
        Returns number of neighbors of process.
        """
        return _mpi.graph_neigh_count(self, rank)

    def Get_neighbors(self, rank):
        """
        Returns list of neighbors of process.
        """
        return _mpi.graph_neigh(self, rank)

    # Graph Low-Level Functions
    # -------------------------
    
    def Map(self, index, edges):
        """
        Returns an optimal placement for the calling process on the
        physical machine.
        """
        return _mpi.graph_map(self, index, edges)

    # Properties
    # ----------
    
    dims = property(Get_dims, doc=None)
    topo = property(Get_topo, doc=None)
    




# --------------------------------------------------------------------
# Intercommunicator
# --------------------------------------------------------------------

class Intercomm(Comm):

    """
    Intercommunicator class.
    """

    def __new__(cls, comm=None):
        return Comm.__new__(cls, comm)

    def __init__(self, comm=None):
        """
        Intercomm initializer.
        """
        Comm.__init__(self, comm)
        if self != _mpi.COMM_NULL:
            if not self.Is_inter():
                raise TypeError('intracommunicator is not allowed')

    # Intercommunicator Accessors
    # ---------------------------

    def Get_remote_size(self):
        """
        Intercommunicator remote size.
        """
        return _mpi.comm_remote_size(self)

    def Get_remote_group(self):
        """
        Accesses the remote group associated with the
        inter-communicator.
        """
        group = _mpi.comm_remote_group(self)
        return Group(group)
        
    # Intercommunicator Constructors
    # ------------------------------

    def Dup(self):
        """
        Duplicate intercommunicator.
        """
        newcomm = _mpi.comm_dup(self)
        return Intercomm(newcomm)
    
    def Create(self, group):
        """
        Create intercommunicator from group.
        """
        newcomm = _mpi.comm_create(self, group)
        return Intercomm(newcomm)
    
    def Split(self, color=0, key=0):
        """
        Split communicator by `color` and `key`.
        """
        newcomm = _mpi.comm_split(self, color, key)
        return Intercomm(newcomm)

    def Merge(self, high=False):
        """
        Merge intercommunicator.
        """
        intracomm = _mpi.intercomm_merge(self, high)
        return Intracomm(intracomm)
    
    # Port
    # ----

    class Port(Comm.Port):
        """Port (Intercommunicator nested class)."""
        pass

    # Sequence Methods
    # ----------------

    def __len__(self):
        """
        Determines the number of ports of a communicator.
        """
        if self == _mpi.COMM_NULL:
            return 0
        else:
            return self.Get_remote_size()

    # Properties
    # ----------
    
    remote_group = property(Get_remote_group, doc=None)
    remote_size  = property(Get_remote_size,  doc=None)





# --------------------------------------------------------------------
# Environmental Management
# --------------------------------------------------------------------

# Initialization and Exit

def Init():
    """
    Initialize the MPI execution environment.
    
    .. note:: `MPI_INITIALIZE` is actually called when this module is
       imported and only if MPI is not already initialized.
    """
    if getattr(Init, '__called', False):
        # this function was already called
        _info = (Init.__module__, Init.__name__)
        raise RuntimeError('%s.%s() already called' % _info)
    else:
        # initialization
        _mpi.init()
        _mpi._set_exception(Exception)
        # register this call
        setattr(Init, '__called', True)
    
def Finalize():
    """
    Terminates MPI execution environment.
    
    .. note:: `MPI_FINALIZE` is actually called if this module had
       initialized MPI.

    .. warning:: ``COMM_WORLD`` and ``COMM_SELF`` will be freed if
       this function is called.
    """
    if getattr(Finalize, '__called', False):
        # this function was already called
        _info = (Init.__module__, Init.__name__)
        raise RuntimeError('%s.%s() already called' % _info)
    else:
        # finalization
        COMM_WORLD.Free()
        COMM_SELF.Free()
        _mpi._set_exception(RuntimeError)
        _mpi.finalize()
        # register this call
        setattr(Finalize, '__called', True)
    
def Is_initialized():
    """
    Indicates whether `MPI_INIT` has been called.
    """
    return _mpi.initialized()

def Is_finalized():
    """
    Indicates whether `MPI_FINALIZE` has completed.

    .. note:: this function does not indicate whether `MPI_INIT` has
      been called.  It only indicates whether `MPI_FINALIZE` has been
      called.
    """
    return _mpi.finalized()

# Implementation Information

# MPI Version Number

def Get_version():
    """
    Obtain the version number of MPI supported by the implementation
    as a tuple ``(version, subversion)``.
    """
    return tuple(_mpi.get_version())

# Environmental Inquires

TAG_UB          = _mpi.TAG_UB
HOST            = _mpi.HOST
IO              = _mpi.IO
WTIME_IS_GLOBAL = _mpi.WTIME_IS_GLOBAL
"""Predefined attributes of `MPI_COMM_WORLD`."""

def Get_processor_name():
    """
    Obtain the name of the calling processor.
    """
    return _mpi.get_processor_name()

# Timers and Synchronization

def Wtime():
    """
    Returns an elapsed time on the calling processor.

    .. note:: This is intended to be a high-resolution, elapsed (or
       wall) clock.  See `MPI_WTICK` to determine the resolution of
       `MPI_WIME` .  If the attribute `MPI_WTIME_IS_GLOBAL` is defined
       and true, then the value is synchronized across all processes
       in `MPI_COMM_WORLD`.
    """
    return _mpi.wtime()

def Wtick():
    """
    Returns the resolution of `MPI_WTIME`.
    """
    return _mpi.wtick()


# Error Handling

class Errhandler(_mpi.Errhandler):

    """
    Errhandler class.
    """
    
    def __new__(cls, errhandler=None):
        return _mpi.Errhandler.__new__(cls, errhandler)

    def __init__(self, errhandler=None):
        _mpi.Errhandler.__init__(self, errhandler)

    def Free(self):
        """
        Frees an error handler.
        """
        _mpi.errhandler_free(self)


# Predefined error handlers

ERRHANDLER_NULL  = Errhandler(_mpi.ERRHANDLER_NULL)
"""Null error handler, `MPI_ERRHANDLER_NULL`."""

ERRORS_ARE_FATAL = Errhandler(_mpi.ERRORS_ARE_FATAL)
ERRORS_RETURN    = Errhandler(_mpi.ERRORS_RETURN)
"""Error-handling specifiers."""


class Exception(RuntimeError):

    """
    Exception class.
    """

    def __init__(self, ierr, *args):
        self.__ierr = 0;
        self.__ierr = int(ierr)
        RuntimeError.__init__(self, int(self), *args)
        
    def __int__(self):
        return self.__ierr
    
    def __str__(self):
        return _mpi.error_string(int(self))

    def Get_error_code(self):
        """
        Error code.
        """
        return int(self)
    
    def Get_error_class(self):
        """
        Error class.
        """
        return _mpi.error_class(int(self))
    
    def Get_error_string(self):
        """
        Error string.
        """
        return _mpi.error_string(int(self))

    ## error_code   = property(Get_error_code)
    ## error_class  = property(Get_error_class)
    ## error_string = property(Get_error_string)

SUCCESS       = _mpi.SUCCESS
ERR_ARG       = _mpi.ERR_ARG
ERR_BUFFER    = _mpi.ERR_BUFFER
ERR_COMM      = _mpi.ERR_COMM
ERR_COUNT     = _mpi.ERR_COUNT
ERR_DIMS      = _mpi.ERR_DIMS
ERR_GROUP     = _mpi.ERR_GROUP
ERR_INTERN    = _mpi.ERR_INTERN
ERR_IN_STATUS = _mpi.ERR_IN_STATUS
ERR_LASTCODE  = _mpi.ERR_LASTCODE
ERR_OP        = _mpi.ERR_OP
ERR_OTHER     = _mpi.ERR_OTHER
ERR_PENDING   = _mpi.ERR_PENDING
ERR_RANK      = _mpi.ERR_RANK
ERR_REQUEST   = _mpi.ERR_REQUEST
ERR_ROOT      = _mpi.ERR_ROOT
ERR_TAG       = _mpi.ERR_TAG
ERR_TOPOLOGY  = _mpi.ERR_TOPOLOGY
ERR_TRUNCATE  = _mpi.ERR_TRUNCATE
ERR_TYPE      = _mpi.ERR_TYPE
ERR_UNKNOWN   = _mpi.ERR_UNKNOWN
"""Constants for MPI error classes."""

def Get_error_class(errorcode):
    """
    Converts an error code into an error class.
    """
    return _mpi.error_class(errorcode)

def Get_error_string(errorcode):
    """
    Returns a string for a given error code.
    """
    return _mpi.error_string(errorcode)



# --------------------------------------------------------------------
# SWIG support
# --------------------------------------------------------------------

def SWIG(cls):
    try:
        import mpi4py._mpi_swig as _swig
        _as   = getattr(_swig, 'as_'  +cls.__name__)
        _from = getattr(_swig, 'from_'+cls.__name__)
        def As_swig(self):
            """Access the object as a *SWIG pointer*."""
            return _as(self)
        def From_swig(cls, obj):
            """Constructs an object from a *SWIG pointer*."""
            return cls(_from(obj))
        this = property(As_swig, doc='SWIG handle')
    except ImportError:
        _swig = None
        def As_swig(self):
            raise ImportError('SWIG support not available')
        def From_swig(cls, obj):
            raise ImportError('SWIG support not available')
        this = property(doc='SWIG handle')
    cls.As_swig   = As_swig
    cls.From_swig = classmethod(From_swig)
    cls.this      = this
    del _swig, As_swig, From_swig, this

map(SWIG, (Op, Comm, Group, Status, Request, Datatype, Errhandler))

del SWIG


# --------------------------------------------------------------------
# Convenience Constants
# --------------------------------------------------------------------

rank = _mpi.rank
"""*rank* of `COMM_WORLD`, shortcut for ``COMM_WORLD.Get_rank()``."""

size = _mpi.size
"""*size* of `COMM_WORLD`, shortcut for ``COMM_WORLD.Get_size()``."""

zero = (rank==0)
"""``True`` if this process has *rank* 0 in ``COMM_WORLD``."""

last = (rank==(size-1))
"""``True`` if this process has *rank* 0 in ``COMM_WORLD``."""

even = ((rank%2)!=1)
"""``True`` if this process has even *rank* in ``COMM_WORLD``."""

odd  = ((rank%2)==1)
"""``True`` if this process has odd *rank* in ``COMM_WORLD``."""





# --------------------------------------------------------------------
# Module initialization
# --------------------------------------------------------------------

def _mpi_init():
    try:
        _mpi._set_exception(Exception)
    except RuntimeError:
        pass
_mpi_init()
del _mpi_init

def _mpi_finalize():
    try:
        _mpi._set_exception(RuntimeError)
        COMM_WORLD.Free()
        COMM_SELF.Free()
    except RuntimeError:
        pass
import atexit
atexit.register(_mpi_finalize)
del atexit, _mpi_finalize




# --------------------------------------------------------------------
# Misc utils
# --------------------------------------------------------------------

def distribute(N, B, i):
    """
    Compute interval [low, high) in bin i
    when sequence of N elements has to be
    distributed over B bins.
    """
    return _mpi.distribute(N, B, i)

def pprint(message = None, root = 0, comm = None):
    """
    Synchronized print.
    """
    inbuf  = _pickle.dump(message)
    if comm is None: comm = WORLD
    result = _mpi.gather_string(inbuf, root, comm)
    if comm.Get_rank() == root:
        for i in result:
            msg = _pickle.load(i)
            print msg
    return

def rprint(message = None, root = 0, comm = None):
    """
    Rooted print.
    """
    if comm is None: comm = WORLD
    if comm.Get_rank() == root:
        print message
    return
