/* $Id: mpi.c,v 1.17 2006/08/03 19:27:30 dalcinl Exp $ */

#define MPICH_IGNORE_CXX_SEEK

#include "libmpi.h"
#include "macros.h"



/*------------------------------------------------------------------*/
/* Module Exception                                                 */
/*------------------------------------------------------------------*/

static PyObject* PyExc_MPIError = NULL;

/*------------------------------------------------------------------*/



/*------------------------------------------------------------------*/
/* Converters                                                       */
/*------------------------------------------------------------------*/

#define MPI4PY_CONVERTER_VAL(name,Type)                      \
static int 						     \
PyObj2##name(PyObject* obj, void* ptr) {		     \
  *((MPI_##Type *)ptr) =  PyMPI##Type##_As##Type(obj);	     \
  return PyErr_Occurred()?0:1;				     \
}							     
#define MPI4PY_CONVERTER_PTR(name,Type)                      \
static int 						     \
PyObj2##name##P(PyObject* obj, void* ptr) {		     \
  *((MPI_##Type **)ptr) =  PyMPI##Type##_As##Type##Ptr(obj); \
  return PyErr_Occurred()?0:1;				     \
}                                                            
#define MPI4PY_CONVERTER_ALL(name,Type)                      \
  MPI4PY_CONVERTER_VAL(name,Type)                            \
  MPI4PY_CONVERTER_PTR(name,Type) 

MPI4PY_CONVERTER_ALL( C , Comm       )
MPI4PY_CONVERTER_ALL( G , Group      )
MPI4PY_CONVERTER_PTR( S , Status     )
MPI4PY_CONVERTER_PTR( R , Request    )
MPI4PY_CONVERTER_ALL( D , Datatype   )
MPI4PY_CONVERTER_ALL( E , Errhandler )




/*------------------------------------------------------------------*/
/* Status                                                            */
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
static
PyObject* 
status_get_count(PyObject* self, PyObject* args) {
  
  MPI_Status*  status;
  MPI_Datatype datatype;

  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2SP, &status,
			 PyObj2D,  &datatype) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
  
  TRY {
    /* get count */
    int count;
    MPI_CHECK_CALL( MPI_Get_count(status, datatype, &count) );
    /* return count */
    return PyInt_FromLong((long)count);
  }
  EXCEPT {
    return NULL;
  }
  
}
/*------------------------------------------------------------------*/
static
PyObject* 
status_get_elems(PyObject* self, PyObject* args) {

  MPI_Status*  status;
  MPI_Datatype datatype;

  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2SP, &status,
			 PyObj2D,  &datatype) ) return NULL;
  
  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);

  TRY {
    /* get count */
    int count;
    MPI_CHECK_CALL( MPI_Get_elements(status, datatype, &count) );
    /* return count */
    return PyInt_FromLong((long)count);
  }
  EXCEPT {
    return NULL;
  }
  
}
/*------------------------------------------------------------------*/
static
PyObject* 
test_cancelled(PyObject* self, PyObject* args) {

  MPI_Status* status;
 
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2SP, &status) ) return NULL;
  
  TRY {
    /* get flag */
    int flag;
    MPI_CHECK_CALL( MPI_Test_cancelled(status, &flag) );
    /* return flag */
    return PyBool_FromLong((long)flag);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/





/*------------------------------------------------------------------*/
/* Request                                                            */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject*
request_wait(PyObject* self, PyObject* args) {

  MPI_Request* request;
  MPI_Status*  status;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2RP, &request,
			 PyObj2SP, &status   ) ) return NULL;

  TRY {
    /* wait */
    MPI_CHECK_CALL( MPI_Wait(request,status) );
    Py_RETURN_NONE;
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
request_test(PyObject* self, PyObject* args) {

  MPI_Request* request;
  MPI_Status*  status;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2RP, &request,
			 PyObj2SP, &status   ) ) return NULL;

  TRY {
    int flag;
    /* test */
    MPI_CHECK_CALL( MPI_Test(request, &flag, status) );
    /* return flag */
    return PyBool_FromLong((long)flag);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
request_free(PyObject* self, PyObject* args) {

  MPI_Request* request;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2RP, &request ) ) return NULL;
  
  MPI_CHECK_NULL(*request, MPI_REQUEST_NULL, MPI_ERR_REQUEST);

  TRY {
    /* free */
    MPI_CHECK_CALL_FREE( MPI_Request_free(request) );
    Py_RETURN_NONE;
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
request_cancel(PyObject* self, PyObject* args) {

  MPI_Request* request;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2RP, &request ) ) return NULL;

  MPI_CHECK_NULL(*request, MPI_REQUEST_NULL, MPI_ERR_REQUEST);
  
  TRY {
    /* cancel */
    MPI_CHECK_CALL( MPI_Cancel(request) );
    Py_RETURN_NONE;
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/






/*------------------------------------------------------------------*/
/* Datatype                                                         */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject*
type_get_extent(PyObject* self, PyObject* args) {
  
  MPI_Datatype datatype;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2D, &datatype ) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
 
  TRY {
    /* get lb and extent */
    MPI_Aint lb;
    MPI_Aint extent;
#if MPI_VERSION == 2
    MPI_CHECK_CALL( MPI_Type_get_extent(datatype, &lb, &extent) );
#else
    MPI_CHECK_CALL( MPI_Type_lb(datatype, &lb) );
    MPI_CHECK_CALL( MPI_Type_extent(datatype, &extent) );
#endif
    /* return lb and extent */
    return Py_BuildValue("NN", PyInt_FromLong((long)lb),
			       PyInt_FromLong((long)extent));
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_extent(PyObject* self, PyObject* args) {
  
  MPI_Datatype datatype;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2D, &datatype ) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
 
  TRY {
    /* get extent */
    MPI_Aint extent;
    MPI_CHECK_CALL( MPI_Type_extent(datatype, &extent) );
    /* return extent */
    return PyInt_FromLong((long)extent);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_lb(PyObject* self, PyObject* args) {
  
  MPI_Datatype datatype;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2D, &datatype ) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
 
  TRY {
    /* get lb */
    MPI_Aint lb;
    MPI_CHECK_CALL( MPI_Type_lb(datatype, &lb) );
    /* return lb */
    return PyInt_FromLong((long)lb);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_ub(PyObject* self, PyObject* args) {
  
  MPI_Datatype datatype;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2D, &datatype ) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
 
  TRY {
    /* get ub */
    MPI_Aint ub;
    MPI_CHECK_CALL( MPI_Type_ub(datatype, &ub) );
    /* return ub */
    return PyInt_FromLong((long)ub);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_size(PyObject* self, PyObject* args) {
  
  MPI_Datatype datatype;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2D, &datatype ) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
 
  TRY {
    /* get size */
    int size;
    MPI_CHECK_CALL( MPI_Type_size(datatype, &size) );
    /* return size */
    return PyInt_FromLong((long)size);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_contiguous(PyObject* self, PyObject* args) {
  
  int count;
  MPI_Datatype datatype;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "iO&",
			 &count,
			 PyObj2D, &datatype ) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
 
  TRY {
    /* get new dattype */
    MPI_Datatype newtype;
    MPI_CHECK_CALL( MPI_Type_contiguous(count, datatype, &newtype) );
    /* return new datatype */
    return PyMPIDatatype_FromDatatype(newtype);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_vector(PyObject* self, PyObject* args) {
  
  int count, blocklength, stride;
  MPI_Datatype datatype;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "iiiO&",
			 &count, &blocklength, &stride,
			 PyObj2D, &datatype ) ) return NULL;

  MPI_CHECK_NULL(datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);
 
  TRY {
    /* get new dattype */
    MPI_Datatype newtype;
    MPI_CHECK_CALL( MPI_Type_vector(count, blocklength, stride,
				    datatype, &newtype) );
    /* return new datatype */
    return PyMPIDatatype_FromDatatype(newtype);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_commit(PyObject* self, PyObject* args) {
  
  MPI_Datatype* datatype;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2DP, &datatype) ) return NULL;

  MPI_CHECK_NULL(*datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);

  TRY {
    /* free datatype */
    MPI_CHECK_CALL( MPI_Type_commit(datatype) );
    Py_RETURN_NONE;
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
type_free(PyObject* self, PyObject* args) {
  
  MPI_Datatype* datatype;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2DP, &datatype) ) return NULL;

  MPI_CHECK_NULL(*datatype, MPI_DATATYPE_NULL, MPI_ERR_TYPE);

  TRY {
    /* free datatype */
    MPI_CHECK_CALL_FREE( MPI_Type_free(datatype) );
    Py_RETURN_NONE;
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/






/*------------------------------------------------------------------*/
/* Group                                                            */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject*
group_size(PyObject* self, PyObject* args) {
  
  MPI_Group group;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2G, &group ) ) return NULL;

  MPI_CHECK_NULL(group, MPI_GROUP_NULL, MPI_ERR_GROUP);
 
  TRY {
    /* get size */
    int size;
    MPI_CHECK_CALL( MPI_Group_size(group, &size) );
    /* return size */
    return PyInt_FromLong((long)size);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_rank(PyObject* self, PyObject* args) {

  MPI_Group group;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2G, &group ) ) return NULL;
  
  MPI_CHECK_NULL(group, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {
    /* get rank */
    int rank;
    MPI_CHECK_CALL( MPI_Group_rank(group,&rank) );
    /* return rank */
    return PyInt_FromLong((long)rank);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_transl_rank(PyObject* self, PyObject* args) {
  
  PyObject* input  = NULL;
  PyObject* result = NULL;

  MPI_Group group1;  
  MPI_Group group2;

  PyObject* s_ranks1 = NULL;
  PyObject* s_ranks2 = NULL;
  int  n      = 0;
  int* ranks1 = NULL;
  int* ranks2 = NULL;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&OO&",
			 PyObj2G, &group1, &input,
			 PyObj2G, &group2) ) return NULL;

  MPI_CHECK_NULL(group1, MPI_GROUP_NULL, MPI_ERR_GROUP);
  MPI_CHECK_NULL(group2, MPI_GROUP_NULL, MPI_ERR_GROUP);
    
  TRY {

    /* handle integer input */
    if ( PyInt_Check(input) ) {
      int rank1 = (int) PyInt_AS_LONG(input);
      int rank2;
      /* translate ranks */
      MPI_CHECK_CALL( MPI_Group_translate_ranks(group1, 1, &rank1,
						group2,    &rank2) );
      return PyInt_FromLong((long)rank2);
    }

    /* handle sequence input */
    s_ranks1 = PySequence_Fast(input,"integer or sequence expected");
    if ( !s_ranks1 ) return NULL;
    n = PySequence_Fast_GET_SIZE(s_ranks1);

    /* allocate ranks* arrays */
    ranks1 = PyMem_New(int, n);
    ranks2 = PyMem_New(int, n);
    if ( !ranks1 || !ranks2 ) PY_RAISE(PyExc_MemoryError, "");

    /* fill ranks1 array */
    { int i; for (i = 0; i < n; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(s_ranks1,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, 
		 "sequence item: integer expected.");
      ranks1[i] = (int) PyInt_AS_LONG(item);
    }}
    
    /*translate ranks */
    MPI_CHECK_CALL( MPI_Group_translate_ranks(group1, n, ranks1,
					      group2,    ranks2) );
    /* allocate and fill result */
    /* output type tries to match input type */
    {
      int tuple = PyTuple_Check(s_ranks1);
      if (tuple) s_ranks2 = PyTuple_New(n);
      else       s_ranks2 =  PyList_New(n);
      
      if ( !s_ranks2 ) RAISE;
      { int i; for (i = 0; i < n; i++) {
	PyObject* item = PyInt_FromLong((long)ranks2[i]);
	if (!item) RAISE;
	if (tuple) PyTuple_SET_ITEM(s_ranks2,i,item);
	else        PyList_SET_ITEM(s_ranks2,i,item);
      }}
      result = s_ranks2; Py_INCREF(s_ranks2);
    }
    
    /* deallocate objects */
    Py_DECREF(s_ranks1);
    Py_DECREF(s_ranks2);
    PyMem_Del(ranks1);
    PyMem_Del(ranks2);

    /* return result */
    return result;

  }

  EXCEPT {
    Py_XDECREF(s_ranks1);
    Py_XDECREF(s_ranks2);
    PyMem_Del(ranks1);
    PyMem_Del(ranks2);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_compare(PyObject* self, PyObject* args) {

  MPI_Group group1;  
  MPI_Group group2;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&", 
			 PyObj2G, &group1,
			 PyObj2G, &group2 ) ) return NULL;

  MPI_CHECK_NULL(group1, MPI_GROUP_NULL, MPI_ERR_GROUP);
  MPI_CHECK_NULL(group2, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {

    /* compare groups */
    int result;
    MPI_CHECK_CALL( MPI_Group_compare(group1,group2,&result) );
    
    /* return result */
    return PyInt_FromLong((long)result);
    
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_group(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {

    MPI_Group newgroup;

    /* get group */
    MPI_CHECK_CALL_COMM(comm,
			MPI_Comm_group(comm, &newgroup) );
    
    /* return group  */
    return PyMPIGroup_FromGroup(newgroup);
    
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_remote_group(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {    

    MPI_Group newgroup;

    /* get group */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Comm_remote_group(comm, &newgroup) );
    
    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT { 
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_union(PyObject* self, PyObject* args) {

  MPI_Group group1;  
  MPI_Group group2;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2G, &group1,
			 PyObj2G, &group2 ) ) return NULL;

  MPI_CHECK_NULL(group1, MPI_GROUP_NULL, MPI_ERR_GROUP);
  MPI_CHECK_NULL(group2, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {
    
    MPI_Group newgroup;

    /* get group */
    MPI_CHECK_CALL( MPI_Group_union(group1, group2, &newgroup) );
    
    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_intersection(PyObject* self, PyObject* args) {

  MPI_Group group1;  
  MPI_Group group2;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2G, &group1,
			 PyObj2G, &group2 ) ) return NULL;


  MPI_CHECK_NULL(group1, MPI_GROUP_NULL, MPI_ERR_GROUP);
  MPI_CHECK_NULL(group2, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {

    MPI_Group newgroup;

    /* get group */
    MPI_CHECK_CALL( MPI_Group_intersection(group1, group2, &newgroup) );
    
    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_difference(PyObject* self, PyObject* args) {

  MPI_Group group1;
  MPI_Group group2;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2G, &group1,
			 PyObj2G, &group2 ) ) return NULL;

  MPI_CHECK_NULL(group1, MPI_GROUP_NULL, MPI_ERR_GROUP);
  MPI_CHECK_NULL(group2, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {

    MPI_Group newgroup;
  
    /* get group */
    MPI_CHECK_CALL( MPI_Group_difference(group1, group2, &newgroup) );
    
    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_incl(PyObject* self, PyObject* args) {

  PyObject* input  = NULL;
  MPI_Group group;

  PyObject* s_ranks = NULL;
  int       n       = 0;
  int*      ranks   = NULL;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O",
			 PyObj2G, &group, &input) ) return NULL;
    
  MPI_CHECK_NULL(group, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {
    
    MPI_Group newgroup;
  
    /* check and handle input objects */
    if ( PyInt_Check(input) ) {
      s_ranks = Py_BuildValue("(i)", (int)PyInt_AS_LONG(input));
      if (s_ranks==NULL) return NULL;
      n = 1;
    }
    else {
      s_ranks = PySequence_Fast(input, "integer or sequence expected");
      if (s_ranks==NULL) return NULL;
      n = PySequence_Fast_GET_SIZE(s_ranks);
    }

    /* allocate ranks array */
    ranks = PyMem_New(int, n);
    if (ranks==NULL) PY_RAISE(PyExc_MemoryError, "");

    /* fill ranks array */
    { int i; for (i = 0; i < n; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(s_ranks,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      ranks[i] = (int)PyInt_AS_LONG(item);
    }}

    /* obtain group */
    MPI_CHECK_CALL( MPI_Group_incl(group, n, ranks, &newgroup) );
    
    /* deallocate objects */
    Py_DECREF(s_ranks);
    PyMem_Del(ranks);

    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT {
    Py_XDECREF(s_ranks);
    PyMem_Del(ranks);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_excl(PyObject* self, PyObject* args) {

  PyObject* input  = NULL;
  MPI_Group group;

  PyObject* s_ranks = NULL;
  int       n       = 0;
  int*      ranks   = NULL;

  MPI_Group newgroup;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O",
			 PyObj2G, &group, &input) ) return NULL;

  MPI_CHECK_NULL(group, MPI_GROUP_NULL, MPI_ERR_GROUP);
  
  TRY {
    
    /* check and handle input objects */
    if ( PyInt_Check(input) ) {
      s_ranks = Py_BuildValue("(i)", (int)PyInt_AS_LONG(input));
      if (s_ranks==NULL) RAISE;
      n = 1;
    }
    else {
      s_ranks = PySequence_Fast(input,"integer or sequence expected");
      if (s_ranks==NULL) RAISE;
      n = PySequence_Fast_GET_SIZE(s_ranks);
    }

    /* allocate ranks array */
    ranks = PyMem_New(int, n);
    if (ranks==NULL) PY_RAISE(PyExc_MemoryError, "");

    /* fill ranks array */
    { int i; for (i = 0; i < n; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(s_ranks,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, 
		 "sequence item: integer expected.");
      ranks[i] = (int) PyInt_AS_LONG(item);
    }}

    /* obtain group */
    MPI_CHECK_CALL( MPI_Group_excl(group, n, ranks, &newgroup) );
    
    /* deallocate objects */
    Py_DECREF(s_ranks);
    PyMem_Del(ranks);
    
    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT {
    Py_XDECREF(s_ranks);
    PyMem_Del(ranks);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_range_incl(PyObject* self, PyObject* args) {

  PyObject* input = NULL;
  MPI_Group group;

  PyObject* s_ranges = NULL;
  PyObject* rseq     = NULL;

  typedef  int range_t[3];
  
  int      n      = 0;
  range_t* ranges = NULL;

  MPI_Group newgroup;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O",
			 PyObj2G, &group, &input) ) return NULL;
  
  MPI_CHECK_NULL(group, MPI_GROUP_NULL, MPI_ERR_GROUP);
    
  TRY {

    s_ranges = PySequence_Fast(input,"sequence expected");
    if (s_ranges==NULL) RAISE;
    n = PySequence_Fast_GET_SIZE(s_ranges);

    /* allocate ranges array */
    ranges = PyMem_New(range_t, n);
    if (ranges==NULL) PY_RAISE(PyExc_MemoryError, "");
    
    /* fill ranges array */
    { int i; for (i = 0; i < n; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(s_ranges,i);
      PyObject* rseq = PySequence_Fast(item,
				       "sequence item: sequence expected");
      if (rseq==NULL) RAISE;
      if (PySequence_Fast_GET_SIZE(rseq) != 3)
	PY_RAISE(PyExc_ValueError, 
		 "sequence item: invalid sequence length");
      { int j; for (j = 0; j < 3; j++) {
	PyObject* r = PySequence_Fast_GET_ITEM(rseq,j);
	if ( !PyInt_Check(r) )
	  PY_RAISE(PyExc_TypeError, 
		   "sequence item: sequence item: integer expected.");
	ranges[i][j] = (int) PyInt_AS_LONG(r);
      }}
      Py_DECREF(rseq); rseq = NULL;
    }}
    
    /* obtain group */
    MPI_CHECK_CALL( MPI_Group_range_incl(group,
					 n, ranges,
					 &newgroup) );
    
    /* deallocate objects */
    PyMem_Del(ranges);
    Py_DECREF(s_ranges);

    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT {
    Py_XDECREF(rseq);
    PyMem_Del(ranges);
    Py_XDECREF(s_ranges);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_range_excl(PyObject* self, PyObject* args) {

  PyObject* input = NULL;
  MPI_Group group;

  PyObject* s_ranges = NULL;
  PyObject* rseq     = NULL;

  typedef  int range_t[3];
  
  int      n      = 0;
  range_t* ranges = NULL;

  MPI_Group newgroup;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O",
			 PyObj2G, &group, &input) ) return NULL;
  
  MPI_CHECK_NULL(group, MPI_GROUP_NULL, MPI_ERR_GROUP);
    
  TRY {

    s_ranges = PySequence_Fast(input,"sequence expected");
    if ( !s_ranges ) RAISE;
    n = PySequence_Fast_GET_SIZE(s_ranges);

    /* allocate ranges array */
    ranges = PyMem_New(range_t, n);
    if ( !ranges ) PY_RAISE(PyExc_MemoryError, "");
    
    /* fill ranges array */
    { int i; for (i = 0; i < n; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(s_ranges,i);
      PyObject* rseq = PySequence_Fast(item,
				       "sequence item: sequence expected");
      if (rseq==NULL) RAISE;
      if (PySequence_Fast_GET_SIZE(rseq) != 3) 
	PY_RAISE(PyExc_ValueError, 
		 "sequence item: invalid sequence length");
      { int j; for (j = 0; j < 3; j++) {
	PyObject* r = PySequence_Fast_GET_ITEM(rseq,j);
	if ( !PyInt_Check(r) )
	  PY_RAISE(PyExc_TypeError, 
		   "sequence item: sequence item: integer expected.");
	ranges[i][j] = (int) PyInt_AS_LONG(r);
      }}
      Py_DECREF(rseq); rseq = NULL;
    }}
    
    /* obtain group */
    MPI_CHECK_CALL( MPI_Group_range_excl(group,
					 n, ranges,
					 &newgroup) );
    
    /* deallocate objects */
    PyMem_Del(ranges);
    Py_DECREF(s_ranges);

    /* return group */
    return PyMPIGroup_FromGroup(newgroup);

  }

  EXCEPT {
    Py_XDECREF(rseq);
    PyMem_Del(ranges);
    Py_XDECREF(s_ranges);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
group_free(PyObject* self, PyObject* args) {
  
  MPI_Group* group;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2GP, &group) ) return NULL;

  MPI_CHECK_NULL(*group, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {

    /* free group */
    MPI_CHECK_CALL_FREE( MPI_Group_free(group) );

    Py_RETURN_NONE;

  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/









/*------------------------------------------------------------------*/
/* Comm - Intracomm                                                 */
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
static 
PyObject*
comm_size(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    /* get size */
    int size;
    MPI_CHECK_CALL_COMM( comm, MPI_Comm_size(comm, &size) );
    /* return size */
    return PyInt_FromLong((long)size);
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_rank(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {  
    /* get rank */
    int rank;
    MPI_CHECK_CALL_COMM( comm, MPI_Comm_rank(comm, &rank) );
    /* return rank */
    return PyInt_FromLong((long)rank);
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_compare(PyObject* self, PyObject* args) {
  
  MPI_Comm comm1; 
  MPI_Comm comm2;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&", 
			 PyObj2C, &comm1,
			 PyObj2C, &comm2 ) ) return NULL;

  MPI_CHECK_NULL(comm1, MPI_COMM_NULL, MPI_ERR_COMM);
  MPI_CHECK_NULL(comm2, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    /* compare communicators */
    int result;
    MPI_CHECK_CALL( MPI_Comm_compare(comm1, comm2, &result) );
    /* return result */
    return PyInt_FromLong((long)result);
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_dup(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    /* duplicate communicator */
    MPI_Comm newcomm;
    MPI_MAKE_NEWCOMM(comm, newcomm,
		     MPI_Comm_dup(comm, &newcomm) );
    /* return result */
    return PyMPIComm_FromComm(newcomm);
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_create(PyObject* self, PyObject* args) {
  
  MPI_Comm  comm;
  MPI_Group group;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&", 
			 PyObj2C, &comm,
			 PyObj2G, &group) ) return NULL;
  
  MPI_CHECK_NULL(comm,  MPI_COMM_NULL,  MPI_ERR_COMM);
  MPI_CHECK_NULL(group, MPI_GROUP_NULL, MPI_ERR_GROUP);

  TRY {    
    /* create communicator */
    MPI_Comm newcomm;
    MPI_MAKE_NEWCOMM(comm, newcomm,
		     MPI_Comm_create(comm, 
				     group, 
				     &newcomm) );
    /* return result */
    return PyMPIComm_FromComm(newcomm);
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_split(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  int color, key;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&ii", 
			 PyObj2C, &comm,
			 &color,&key) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    /* split communicator */
    MPI_Comm newcomm;
    MPI_MAKE_NEWCOMM(comm, newcomm,
		     MPI_Comm_split(comm, 
				    color, key, 
				    &newcomm) );
    /* return new communicator */
    return PyMPIComm_FromComm(newcomm);
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_free(PyObject* self, PyObject* args) {
  
  MPI_Comm* comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2CP, &comm ) ) return NULL;
  
  MPI_CHECK_NULL(*comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    /* free communicator */
    MPI_CHECK_CALL_FREE( MPI_Comm_free(comm) );
    Py_RETURN_NONE;
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/










/*------------------------------------------------------------------*/
/* Intercomm                                                        */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject*
comm_test_inter(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {

    /* test */
    int flag;
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Comm_test_inter(comm, &flag) );
    
    /* return result */
    return PyBool_FromLong((long)flag);
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
comm_remote_size(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    
    /* get size */
    int size;
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Comm_remote_size(comm,&size) );
    
    /* return size */
    return PyInt_FromLong((long)size);
    
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
intercomm_create(PyObject* self, PyObject* args) {

  MPI_Comm local_comm; 
  MPI_Comm bridge_comm;
  int local_leader, remote_leader, tag;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&iO&ii",
			 PyObj2C, &local_comm,  &local_leader,
			 PyObj2C, &bridge_comm, &remote_leader,
			 &tag) ) return NULL;

  MPI_CHECK_NULL(local_comm,  MPI_COMM_NULL, MPI_ERR_COMM);
  MPI_CHECK_NULL(bridge_comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    
    /* create intercommunicator */
    MPI_Comm newintercomm;
    MPI_MAKE_NEWCOMM(local_comm, newintercomm,
		     MPI_Intercomm_create(local_comm,
					  local_leader,
					  bridge_comm, 
					  remote_leader, 
					  tag, 
					  &newintercomm) );

    /* return new communicator */
    return PyMPIComm_FromComm(newintercomm);
    
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
intercomm_merge(PyObject* self, PyObject* args) {
  
  MPI_Comm comm; 
  int high;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&i",
			 PyObj2C, &comm,
			 &high) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    
    /* merge intercommunicator */
    MPI_Comm newintracomm;
    MPI_MAKE_NEWCOMM(comm, newintracomm,
		     MPI_Intercomm_merge(comm, 
					 high,
					 &newintracomm) );

    /* return new communicator */
    return PyMPIComm_FromComm(newintracomm);
    
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/










/*------------------------------------------------------------------*/
/* Virtual Topologies                                               */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject*
topo_test(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    
    /* test communicator */
    int status;
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Topo_test(comm, &status) );
    
    /* return result */
    return PyInt_FromLong((long)status);
  }
  
  EXCEPT {
    return NULL;
  }

}


/*------------------------------------------------------------------*/
/* Cartcomm                                                         */
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
static 
PyObject*
cart_create(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  PyObject* o_dims = NULL;
  PyObject* o_pers = NULL;
  PyObject* o_reor = NULL;

  int  ndims   = 0;
  int* dims    = NULL;
  int* periods = NULL;
  int  reorder = 0;

  MPI_Comm comm_cart;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&OOO",
			 PyObj2C, &comm,
			 &o_dims, &o_pers, &o_reor)) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    /* check input objects */
    o_dims = PySequence_Fast(o_dims,"sequence expected");
    if ( !o_dims ) RAISE;
    o_pers = PySequence_Fast(o_pers,"sequence expected");
    if ( !o_pers ) RAISE;
    reorder = PyObject_IsTrue(o_reor);
    if ( reorder == -1 )
      PY_RAISE(PyExc_TypeError, "boolean expected");

    /* allocate arrays       */
    ndims = PySequence_Fast_GET_SIZE(o_dims);
    if ( ndims != PySequence_Fast_GET_SIZE(o_pers) )
      PY_RAISE(PyExc_ValueError, "incompatible sequence sizes");
    dims    = PyMem_New(int, ndims);
    periods = PyMem_New(int, ndims);
    if ( !dims || !periods) PY_RAISE(PyExc_MemoryError, "");

    /* fill arrays */
    { int i; for (i = 0; i < ndims; i++) {
      PyObject* d = PySequence_Fast_GET_ITEM(o_dims,i);
      PyObject* p = PySequence_Fast_GET_ITEM(o_pers,i);
      if ( !PyInt_Check(d) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      dims[i] = (int) PyInt_AS_LONG(d);
      periods[i] = PyObject_IsTrue(p);
      if ( periods[i] == -1 )
	PY_RAISE(PyExc_TypeError, "sequence item: boolean expected.");
    }}

    /* create cartesian communicator */
    MPI_MAKE_NEWCOMM(comm, comm_cart,
		     MPI_Cart_create(comm, 
				     ndims, dims, 
				     periods, reorder,
				     &comm_cart) );

    /* deallocate objects */
    PyMem_Del(dims);
    PyMem_Del(periods);
    Py_DECREF(o_dims);
    Py_DECREF(o_pers);

    /* return new communicator */
    return PyMPIComm_FromComm(comm_cart);

  }

  EXCEPT {
    PyMem_Del(dims);
    PyMem_Del(periods);
    Py_XDECREF(o_dims);
    Py_XDECREF(o_pers);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
dims_create(PyObject* self, PyObject* args) {
  
  int nnodes;
  PyObject* input  = NULL;
  PyObject* result = NULL;

  int       ndims  = 0;
  int*      dims   = NULL;
  PyObject* s_dims = NULL;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "iO",
			 &nnodes, &input) ) return NULL;

  TRY {
    /* check input objects */
    if ( PyInt_Check(input) ) {
      ndims = (int) PyInt_AS_LONG(input);
    }
    else {
      s_dims = PySequence_Fast(input,"integer or sequence expected");
      if ( !s_dims ) return NULL;
      ndims = PySequence_Fast_GET_SIZE(s_dims);
    }

    /* allocate dims array */
    dims  = PyMem_New(int, ndims);
    if ( !dims ) PY_RAISE(PyExc_MemoryError, "");

    /* fill dims array */
    if (!s_dims)
      { int i; for (i = 0; i < ndims; i++) 
	dims[i] = 0;
      }
    else
      { int i; for (i = 0; i < ndims; i++) {
	PyObject* item = PySequence_Fast_GET_ITEM(s_dims,i);
	if ( !PyInt_Check(item) )
	  PY_RAISE(PyExc_TypeError,
		   "sequence item: integer expected.");
	dims[i] = (int) PyInt_AS_LONG(item);
      }}

    /* obtain dims array */
    MPI_CHECK_CALL( MPI_Dims_create(nnodes, ndims, dims) );
    
    /* construct result */
    result = PyTuple_New(ndims);
    if ( !result ) RAISE;
    { int i; for (i = 0; i < ndims; i++) {
      PyObject* o = PyInt_FromLong((long)dims[i]);
      if ( !o ) RAISE;
      PyTuple_SET_ITEM(result, i, o);
    }}

    /* deallocate objects */
    Py_XDECREF(s_dims);
    PyMem_Del(dims); dims = NULL;
    
    /* return result */
    return result;

  }

  EXCEPT {
    PyMem_Del(dims);
    Py_XDECREF(s_dims);
    Py_XDECREF(result);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
cartdim_get(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2C, &comm) ) return NULL;
  

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    /* obtain ndims */
    int  ndims;
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cartdim_get(comm, &ndims) );
    
    /* return result */
    return PyInt_FromLong((long)ndims);
    
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
cart_get(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  PyObject* o_dims    = NULL;
  PyObject* o_periods = NULL;
  PyObject* o_coords  = NULL;
  
  int  maxdims;
  int* dims    = NULL;
  int* periods = NULL;
  int* coords  = NULL;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    
    /* obtain maxdims */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cartdim_get(comm, &maxdims) );

    /* allocate arrays */
    dims    = PyMem_New(int, maxdims);
    periods = PyMem_New(int, maxdims);
    coords  = PyMem_New(int, maxdims);
    if ( !dims || !periods || !coords ) 
      PY_RAISE(PyExc_MemoryError, "");
    
    /* get array's data */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cart_get(comm, maxdims, 
				     dims, periods, coords) );

    /* allocate result lists */
    o_dims    = PyTuple_New(maxdims);
    o_periods = PyTuple_New(maxdims);
    o_coords  = PyTuple_New(maxdims);
    if ( !o_dims || !o_periods || !o_coords ) RAISE;

    /* fill result lists */
    { int i; for (i = 0; i < maxdims; i++){
      PyObject* item;
      /* dims */
      item = PyInt_FromLong((long)dims[i]);
      if (!item) RAISE;
      PyTuple_SET_ITEM(o_dims,i,item);
      /* periods */
      item = PyInt_FromLong((long)periods[i]);
      if (!item) RAISE;
      PyTuple_SET_ITEM(o_periods,i,item);
      /* coords */
      item = PyInt_FromLong((long)coords[i]);
      if (!item) RAISE;
      PyTuple_SET_ITEM(o_coords,i,item);
    }}

    /* deallocate arrays */
    PyMem_Del(dims);
    PyMem_Del(periods);
    PyMem_Del(coords);

    /* return result */
    return Py_BuildValue("NNN", o_dims, o_periods, o_coords);
    
  }

  EXCEPT {
    PyMem_Del(dims);  
    PyMem_Del(periods);
    PyMem_Del(coords);
    Py_XDECREF(o_dims);
    Py_XDECREF(o_periods);
    Py_XDECREF(o_coords);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
cart_rank(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;
  PyObject* o_coords = NULL;

  int* coords = NULL;
  int  rank;
  int  ndims;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O",
			 PyObj2C, &comm, &o_coords) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    
    /* check input arguments */
    o_coords = PySequence_Fast(o_coords,"sequence expected");
    if ( !o_coords ) RAISE;

    /* allocate coords array */
    ndims = PySequence_Fast_GET_SIZE(o_coords);
    coords = PyMem_New(int, ndims);
    if ( !coords ) PY_RAISE(PyExc_MemoryError, "");
    
    /* fill coords array */
    { int i; for (i = 0; i < ndims; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(o_coords,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      coords[i] = (int) PyInt_AS_LONG(item);
    }}
    
    /* obtain rank */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cart_rank(comm, coords, &rank) );

    /* deallocate objects */
    PyMem_Del(coords);
    Py_DECREF(o_coords);
    /* return result */
    return PyInt_FromLong((long)rank);
  }

  EXCEPT {
    PyMem_Del(coords);
    Py_XDECREF(o_coords);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
cart_coords(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;

  int  rank;
  int  maxdims;
  int* coords = NULL;

  PyObject* result = NULL;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&i",
			 PyObj2C, &comm, &rank) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    
    /* allocate coords array */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cartdim_get(comm,&maxdims) );
    coords = PyMem_New(int, maxdims);
    if ( !coords ) PY_RAISE(PyExc_MemoryError, "");

    /* obtain coords */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cart_coords(comm, rank, maxdims, coords) );
    
    /* construct result */
    result = PyTuple_New(maxdims);
    if ( !result ) RAISE;
    { int i; for (i = 0; i < maxdims; i++) {
      PyObject* item = PyInt_FromLong((long) coords[i]);
      if ( !item ) PY_RAISE(PyExc_MemoryError, "");
      PyTuple_SET_ITEM(result, i, item);
    }}

    /* deallocate objects */
    PyMem_Del(coords);
    
    /* return result */
    return result;

  }

  EXCEPT {
    PyMem_Del(coords);
    Py_XDECREF(result);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
cart_shift(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  int direction;
  int disp;

  int rank_source;
  int rank_dest;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&ii",
			 PyObj2C, &comm,
			 &direction, &disp) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    /* obtain source and dest */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cart_shift(comm, direction, disp,
				       &rank_source, &rank_dest) );

    /* return result */
    return Py_BuildValue("ii",rank_source,rank_dest);

  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
cart_sub(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  PyObject* o_rem_dims = NULL;

  int  ndims;
  int* remain_dims = NULL;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O",
			 PyObj2C, &comm, &o_rem_dims) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
    
  TRY {

    MPI_Comm newcomm;

    /* check input objects */
    o_rem_dims = PySequence_Fast(o_rem_dims, "sequence expected");
    if ( !o_rem_dims ) RAISE;
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Cartdim_get(comm,&ndims) );
    if ( ndims != PySequence_Fast_GET_SIZE(o_rem_dims) )
      PY_RAISE(PyExc_ValueError, "invalid sequence size");

    /* allocate array */
    remain_dims = PyMem_New(int, ndims);
    if ( !remain_dims ) PY_RAISE(PyExc_MemoryError, "");
    
    /* fill array */
    { int i; for (i = 0; i < ndims; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(o_rem_dims,i);
      remain_dims[i] = PyObject_IsTrue(item);
      if ( remain_dims[i] == -1 )
	PY_RAISE(PyExc_TypeError, "sequence item: boolean expected.");
    }}

    /* create cartesian communicator */
    MPI_MAKE_NEWCOMM(comm, newcomm,
		     MPI_Cart_sub(comm, 
				  remain_dims, 
				  &newcomm) );
    
    /* deallocate objects */
    PyMem_Del(remain_dims);
    Py_DECREF(o_rem_dims);
    
    /* return new communicator */
    return PyMPIComm_FromComm(newcomm);

  }

   EXCEPT {
    PyMem_Del(remain_dims);
    Py_XDECREF(o_rem_dims);
    return NULL;
  }
  
}
/*------------------------------------------------------------------*/
static 
PyObject*
cart_map(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  PyObject* o_dims = NULL;
  PyObject* o_pers = NULL;

  int  ndims   = 0;
  int* dims    = NULL;
  int* periods = NULL;

  int newrank;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&OO",
			 PyObj2C, &comm,
			 &o_dims, &o_pers)) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    /* check input objects */
    o_dims = PySequence_Fast(o_dims,"sequence expected");
    if ( !o_dims ) RAISE;
    o_pers = PySequence_Fast(o_pers,"sequence expected");
    if ( !o_pers ) RAISE;

    /* allocate arrays       */
    ndims = PySequence_Fast_GET_SIZE(o_dims);
    if ( ndims != PySequence_Fast_GET_SIZE(o_pers) )
      PY_RAISE(PyExc_ValueError, "incompatible sequence sizes");
    dims    = PyMem_New(int, ndims);
    periods = PyMem_New(int, ndims);
    if ( !dims || !periods) PY_RAISE(PyExc_MemoryError, "");

    /* fill arrays */
    { int i; for (i = 0; i < ndims; i++) {
      PyObject* d = PySequence_Fast_GET_ITEM(o_dims,i);
      PyObject* p = PySequence_Fast_GET_ITEM(o_pers,i);
      if ( !PyInt_Check(d) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      dims[i] = (int) PyInt_AS_LONG(d);
      periods[i] = PyObject_IsTrue(p);
      if ( periods[i] == -1 )
	PY_RAISE(PyExc_TypeError, "sequence item: boolean expected.");
    }}

    /* create cartesian communicator */
    MPI_CHECK_CALL_COMM(comm,
			MPI_Cart_map(comm,
				     ndims, dims, periods,
				     &newrank) );

    /* deallocate objects */
    PyMem_Del(dims);
    PyMem_Del(periods);
    Py_DECREF(o_dims);
    Py_DECREF(o_pers);

    /* return reordered rank */
    return Py_BuildValue("i",newrank);

  }

  EXCEPT {
    PyMem_Del(dims);
    PyMem_Del(periods);
    Py_XDECREF(o_dims);
    Py_XDECREF(o_pers);
    return NULL;
  }

}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Graphcomm                                                        */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static
PyObject*
graph_create(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  PyObject* o_index = NULL;
  PyObject* o_edges = NULL;
  PyObject* o_reord = NULL;

  int  nnodes;
  int* index = NULL;
  int  nedges;
  int* edges = NULL;
  int  reorder;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&OOO",
			 PyObj2C, &comm,
			 &o_index, &o_edges, &o_reord)) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    MPI_Comm comm_graph;

    /* check input objects */
    o_index = PySequence_Fast(o_index, "sequence expected");
    if ( !o_index ) RAISE;
    o_edges = PySequence_Fast(o_edges, "sequence expected");
    if ( !o_edges ) RAISE;
    reorder = PyObject_IsTrue(o_reord);
    if ( reorder == -1 )
      PY_RAISE(PyExc_TypeError, "boolean expected");

    /* allocate arrays */
    nnodes = PySequence_Fast_GET_SIZE(o_index);
    nedges = PySequence_Fast_GET_SIZE(o_edges);
    index = PyMem_New(int, nnodes);
    edges = PyMem_New(int, nedges);
    if ( !index || !edges ) PY_RAISE(PyExc_MemoryError, "");

    /* fill arrays */
    { int i; for (i = 0; i < nnodes; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(o_index,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      index[i] = (int) PyInt_AS_LONG(item);
    }}
    { int i; for (i = 0; i < nedges; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(o_edges,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      edges[i] = (int) PyInt_AS_LONG(item);
    }}

    /* create graph communicator */
    MPI_MAKE_NEWCOMM(comm, comm_graph,
		     MPI_Graph_create(comm, 
				      nnodes, index, edges, reorder, 
				      &comm_graph) );

    /* deallocate objects */
    PyMem_Del(index);
    PyMem_Del(edges);
    Py_DECREF(o_index);
    Py_DECREF(o_edges);

    /* return new communicator */
    return PyMPIComm_FromComm(comm_graph);
  }

  EXCEPT {
    PyMem_Del(index);
    PyMem_Del(edges);
    Py_XDECREF(o_index);
    Py_XDECREF(o_edges);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
graphdims_get(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  
  int nnodes;
  int nedges;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  
  TRY {
    
    /* obtain nnodes and nedges */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Graphdims_get(comm, &nnodes, &nedges) );

    /* return result */
    return Py_BuildValue("ii", nnodes, nedges);
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
graph_get(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  /* parse input arguments */

  int  maxindex;
  int  maxedges;
  int* index = NULL;
  int* edges = NULL;
  
  PyObject* o_index = NULL;
  PyObject* o_edges = NULL;

  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
    

  TRY {

    /* obtain maxindex and maxedges */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Graphdims_get(comm, &maxindex, &maxedges) );

    /* allocate arrays */
    index = PyMem_New(int, maxindex);
    edges = PyMem_New(int, maxedges);
    if ( !index || !edges ) PY_RAISE(PyExc_MemoryError, "");
    
    /* get array's data */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Graph_get(comm, 
				      maxindex, maxedges,
				      index,    edges) );

    /* allocate result lists */
    o_index = PyList_New(maxindex);
    o_edges = PyList_New(maxedges);
    if ( !o_index || !o_edges ) RAISE;

    /* fill result lists */
    { int i; for (i = 0; i < maxindex; i++) {
      PyObject* item = PyInt_FromLong((long)index[i]);
      if (!item) RAISE;
      PyList_SET_ITEM(o_index,i,item);
    }}
    { int i; for (i = 0; i < maxedges; i++) {
      PyObject* item = PyInt_FromLong((long)edges[i]);
      if (!item) RAISE;
      PyList_SET_ITEM(o_edges,i,item);
    }}

    /* deallocate arrays */
    PyMem_Del(index);
    PyMem_Del(edges);

    /* return result */
    return Py_BuildValue("NN", o_index, o_edges);

  }
  
  EXCEPT {
    PyMem_Del(index);
    PyMem_Del(edges);
    Py_XDECREF(o_index);
    Py_XDECREF(o_edges);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
graph_neigh_count(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  int rank;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&i",
			 PyObj2C, &comm, &rank) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    int nneighbors;
   
    /* obtain nnodes and nedges */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Graph_neighbors_count(comm, rank,
						  &nneighbors) );
    /* return result */
    return PyInt_FromLong((long)nneighbors);
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
graph_neigh(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  int rank;

  int  maxneighbors;
  int* neighbors = NULL;

  PyObject* o_neighbors = NULL;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&i",
			 PyObj2C, &comm, &rank) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
 
  TRY {
    
    /* obtain maxneighbors */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Graph_neighbors_count(comm, rank, &maxneighbors) );
    
    /* allocate array */
    neighbors = PyMem_New(int, maxneighbors);
    if ( !neighbors ) PY_RAISE(PyExc_MemoryError, "");
    
    /* get array data */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Graph_neighbors(comm, rank,
					    maxneighbors, neighbors) );
    
    /* allocate result list */
    o_neighbors = PyList_New(maxneighbors);
    if ( !o_neighbors ) RAISE;
    
    /* fill list */
    { int i; for (i = 0; i < maxneighbors; i++){
      PyObject* item = PyInt_FromLong((long)neighbors[i]);
      if (!item) PY_RAISE(PyExc_MemoryError, "");
      PyList_SET_ITEM(o_neighbors,i,item);
    }}

    /* deallocate array */
    PyMem_Del(neighbors);

    /* return result */
    return o_neighbors;
  }

  EXCEPT {
    PyMem_Del(neighbors);
    Py_XDECREF(o_neighbors);
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject*
graph_map(PyObject* self, PyObject* args) {

  MPI_Comm comm;

  PyObject* o_index = NULL;
  PyObject* o_edges = NULL;

  int  nnodes;
  int* index = NULL;
  int  nedges;
  int* edges = NULL;

  int  newrank;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&OO",
			 PyObj2C, &comm,
			 &o_index, &o_edges)) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    /* check input objects */
    o_index = PySequence_Fast(o_index, "sequence expected");
    if ( !o_index ) RAISE;
    o_edges = PySequence_Fast(o_edges, "sequence expected");
    if ( !o_edges ) RAISE;

    /* allocate arrays */
    nnodes = PySequence_Fast_GET_SIZE(o_index);
    nedges = PySequence_Fast_GET_SIZE(o_edges);
    index = PyMem_New(int, nnodes);
    edges = PyMem_New(int, nedges);
    if ( !index || !edges ) PY_RAISE(PyExc_MemoryError, "");

    /* fill arrays */
    { int i; for (i = 0; i < nnodes; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(o_index,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      index[i] = (int) PyInt_AS_LONG(item);
    }}
    { int i; for (i = 0; i < nedges; i++) {
      PyObject* item = PySequence_Fast_GET_ITEM(o_edges,i);
      if ( !PyInt_Check(item) )
	PY_RAISE(PyExc_TypeError, "sequence item: integer expected.");
      edges[i] = (int) PyInt_AS_LONG(item);
    }}

    /* create graph communicator */
    MPI_CHECK_CALL_COMM(comm,
			MPI_Graph_map(comm,
				      nnodes, index, edges,
				      &newrank) );

    /* deallocate objects */
    PyMem_Del(index);
    PyMem_Del(edges);
    Py_DECREF(o_index);
    Py_DECREF(o_edges);

    /* return reorder rank */
    return Py_BuildValue("i",newrank);
  }

  EXCEPT {
    PyMem_Del(index);
    PyMem_Del(edges);
    Py_XDECREF(o_index);
    Py_XDECREF(o_edges);
    return NULL;
  }

}
/*------------------------------------------------------------------*/







#define PICKLED_STRING MPI_CHAR


/*------------------------------------------------------------------*/
/* Point-to-Point Communications                                    */
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Send                                                             */
/*------------------------------------------------------------------*/

static 
PyObject*
send_string(PyObject* self, PyObject* args) {
  
  PyObject* input  = NULL;
  
  int dest, tag;  
  MPI_Comm comm;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "SiiO&", 
			 &input,
			 &dest, &tag, 
			 PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    int   count;
    char* buffer;
    
    /* send string size */
    count = PyString_GET_SIZE(input);
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Send(&count, 1, MPI_INT, dest, tag, comm) );
    
    /* send string data */
    buffer = PyString_AS_STRING(input);
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Send(buffer, count, PICKLED_STRING, 
				 dest, tag, comm) );

    Py_RETURN_NONE;
    
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Recv                                                             */
/*------------------------------------------------------------------*/
static 
PyObject* 
recv_string(PyObject* self, PyObject* args) {

  PyObject* result = NULL;

  int source, tag;
  MPI_Comm    comm;
  MPI_Status* status;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "iiO&O&", 
			 &source, &tag, 
			 PyObj2C, &comm,
			 PyObj2SP, &status) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    int   count;
    char* buffer;
    
    /* recv string size */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Recv(&count, 1, MPI_INT,
				 source, tag, comm, status) );
    
    /* create python string */
    result = PyString_FromStringAndSize(NULL, count);
    if (!result) RAISE;
    
    /* receive string data */
    buffer = PyString_AS_STRING(result);
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Recv(buffer, count, PICKLED_STRING,
				 source, tag, comm, status) );
    
    /* return result */
    return result;
  }
  
  
  EXCEPT {
    Py_XDECREF(result);
    return NULL;
  }

}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Send-Recv                                                        */
/*------------------------------------------------------------------*/
static 
PyObject*
sendrecv_string(PyObject* self, PyObject* args) {
  
  PyObject* input  = NULL;
  PyObject* result = NULL;

  int dest,    source;
  int sendtag, recvtag;
  MPI_Comm    comm;
  MPI_Status* status;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "SiiiiO&O&", 
			 &input,
			 &dest,    &source,
			 &sendtag, &recvtag,
			 PyObj2C, &comm,
			 PyObj2SP, &status) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
    
  TRY {

    int   scount;
    int   rcount;
    char* sbuff;
    char* rbuff;
    
    /* sendrecv string size */
    scount = PyString_GET_SIZE(input);
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Sendrecv(&scount, 1, MPI_INT, dest,   sendtag,
				     &rcount, 1, MPI_INT, source, recvtag,
				     comm, status) );
    
    /* create python string */
    result = PyString_FromStringAndSize(NULL, rcount);
    if (!result) RAISE;
    
    /* sendrecv string data */
    sbuff = PyString_AS_STRING(input);
    rbuff = PyString_AS_STRING(result);
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Sendrecv(sbuff, scount, PICKLED_STRING, 
				     dest,   sendtag,
				     rbuff, rcount, PICKLED_STRING, 
				     source, recvtag,
				     comm, status) );
    
    /* return result */
    return result;

  }
  
  
  EXCEPT {
    Py_XDECREF(result);
    return NULL;
  }
  
}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Probe                                                            */
/*------------------------------------------------------------------*/
static 
PyObject*
comm_probe(PyObject* self, PyObject* args) {

  int source, tag;
  MPI_Comm    comm;
  MPI_Status* status;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "iiO&O&",
			 &source, &tag,
			 PyObj2C,  &comm,
			 PyObj2SP, &status) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Probe(source, tag, comm, status) );
    
    Py_RETURN_NONE;
    
  }  
  
  EXCEPT {
    return NULL;
  }
 
}


/*------------------------------------------------------------------*/









/*------------------------------------------------------------------*/
/* Collective Communications                                        */
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Barrier                                                          */
/*------------------------------------------------------------------*/
static 
PyObject* 
barrier(PyObject* self, PyObject* args) {
  
  MPI_Comm comm;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2C, &comm ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    /* barrier synchronization */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Barrier(comm) );
    Py_RETURN_NONE;
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Bcast                                                            */
/*------------------------------------------------------------------*/
static 
PyObject* 
bcast_string(PyObject* self, PyObject* args) {

  PyObject* input  = NULL;
  PyObject* result = NULL;
  
  int root; 
  MPI_Comm comm;

  int   count  = 0;
  char* buffer = NULL;

  /* parse input arguments */
  if (!PyArg_ParseTuple(args, "OiO&",
			&input, 
			&root, PyObj2C, &comm)) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    /* comm rank */
    int rank;
    MPI_CHECK( MPI_Comm_rank(comm,&rank) );
   
    
    /* ---------   */
    /* root code */
    /* --------- */
    if (rank == root ) {
      
      /* check input object */
      if (PyString_Check(input)) 
	count = PyString_GET_SIZE(input);
      else
	count = -1;
      
      /* bcast string size */
      MPI_CHECK_CALL_COMM(comm, 
			  MPI_Bcast(&count, 1, MPI_INT, root, comm) );

      if (count == -1) PY_RAISE(PyExc_TypeError, "expecting a string.");
      
      Py_INCREF(input); /* bcast is already */
      result = input;   /* IN PLACE on root */
   
    }
    

    /* ----------- */
    /* others code */
    /* ----------- */
    else { 
      
      /* receive string size */
      MPI_CHECK_CALL_COMM(comm, 
			  MPI_Bcast(&count, 1, MPI_INT, root, comm) );

      if (count == -1) PY_RAISE(PyExc_RuntimeError, "in root processor");

      /* create python string */
      result = PyString_FromStringAndSize(NULL, count);
      if (!result) RAISE;
      
    }

    /* ---------   */
    /* all code */
    /* --------- */
    
    /* bcast string buffer */
    buffer = PyString_AS_STRING(result);
    MPI_CHECK_CALL_COMM(comm,
			MPI_Bcast(buffer, count, PICKLED_STRING, 
				  root, comm) );
    
    /* return string */
    return result;
    
  }
  
  EXCEPT {
    Py_XDECREF(result);
    return NULL;
  }

}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Gather                                                           */
/*------------------------------------------------------------------*/
static 
PyObject* 
gather_string(PyObject* self, PyObject* args) {
  
  PyObject* input  = NULL;
  PyObject* result = NULL;

  int root; 
  MPI_Comm comm;
  
  char* sbuff   = NULL;
  int   scount  = 0;
  
  char* rbuff   = NULL;
  int*  rcounts = NULL;
  int*  displs  = NULL; 

  /* parse input arguments */
  if (!PyArg_ParseTuple(args, "OiO&",
			&input, 
			&root, PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
    
  TRY {

    /* ---------   */
    /* all code */
    /* --------- */

    /* comm rank */
    int rank;
    MPI_CHECK( MPI_Comm_rank(comm,&rank) );

    /* check input object */
    if (!PyString_Check(input))
      PY_RAISE(PyExc_TypeError, "expecting a string.");
    sbuff   = PyString_AS_STRING(input);
    scount = PyString_GET_SIZE(input);
    
    
    /* ---------   */
    /* root code */
    /* --------- */
    if (rank == root ) {
      
      /* comm size */
      int size;
      MPI_CHECK( MPI_Comm_size(comm, &size) );

      /* allocate arrays 'rcounts' and 'displs' */
      rcounts = PyMem_New(int, size);
      displs     = PyMem_New(int, size);
      if ( !rcounts || !displs )
	PY_RAISE(PyExc_MemoryError, "");
      
      /* gather string sizes */
      MPI_CHECK_CALL_COMM(comm, 
			  MPI_Gather(&scount, 1, MPI_INT,
				     rcounts, 1, MPI_INT,
				     root, comm) );
      
      /* create result list */
      result = PyList_New(size);
      if (!result) RAISE;
      
      /* fill result list */
      { int i; for (i = 0; i < size; i++) {
	/* create python string */
	PyObject* s = PyString_FromStringAndSize(NULL, rcounts[i]);
	if ( !s ) RAISE;
	/* set string in result list */
	PyList_SET_ITEM(result,i,s);
	/* calculate 'rbuff' pointer and 'displs' array */
	if (i == 0) rbuff = PyString_AS_STRING(s);
	displs[i] = PyString_AS_STRING(s) - rbuff;
      }}
      
      /* gather string buffers */
      MPI_CHECK_CALL_COMM(comm, 
			  MPI_Gatherv(sbuff, scount,          PICKLED_STRING,
				      rbuff, rcounts, displs, PICKLED_STRING,
				      root,comm) );
      
      /* deallocate memory */
      PyMem_Del(rcounts);
      PyMem_Del(displs);

      /* return result list */
      return  result;
      
    }
    

    /* ----------- */
    /* others code */
    /* ----------- */
    else { 
      
      /* send string size */
      MPI_CHECK_CALL_COMM(comm,
			  MPI_Gather(&scount, 1, MPI_INT,
				     rcounts, 1, MPI_INT,
				     root, comm) );
      
      /* send string buffer */
      MPI_CHECK_CALL_COMM(comm,
			  MPI_Gatherv(sbuff, scount,          PICKLED_STRING,
				      rbuff, rcounts, displs, PICKLED_STRING,
				      root,comm) );
      
      Py_RETURN_NONE;
      
    }
    
  }
  
  EXCEPT {
    PyMem_Del(rcounts);
    PyMem_Del(displs);
    Py_XDECREF(result);
    return NULL;
  }

}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Scatter                                                          */
/*------------------------------------------------------------------*/
static 
PyObject* 
scatter_string(PyObject* self, PyObject* args) {

  PyObject* input  = NULL;
  PyObject* result = NULL;

  int root; 
  MPI_Comm comm;
  
  char* sbuff   = NULL;
  int*  scounts = NULL;
  int*  displs  = NULL;
  
  char* rbuff   = NULL;
  int   rcount  = 0;

  /* parse input arguments */
  if (!PyArg_ParseTuple(args, "OiO&",
			&input,
			&root, PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {

    /* ---------   */
    /* all code */
    /* --------- */

    /* comm rank */
    int rank, size;
    MPI_CHECK( MPI_Comm_rank(comm,&rank) );
    MPI_CHECK( MPI_Comm_size(comm,&size) );
    
    
    /* ---------   */
    /* root code */
    /* --------- */
    if (rank == root ) {
      
      /* allocate arrays 'scounts' and 'displs' */
      scounts = PyMem_New(int, size);
      displs     = PyMem_New(int, size);
      if ( !scounts || !displs )
	PY_RAISE(PyExc_MemoryError, "");

      /* check input object */
      if ( !PyList_Check(input) ) {
	PY_SET_ERROR(PyExc_TypeError, "expecting a list.");
	goto error;
      }
      /* check list length & comm size */
      if (size != PyList_GET_SIZE(input) ) {
	PY_SET_ERROR(PyExc_ValueError, "invalid list length.");
	goto error;
      }
      
      /* calculate 'sbuff', 'scounts' and 'displs' */
      { int i; for (i = 0; i < size; i++) {
	PyObject* s = PyList_GET_ITEM(input, i);
	if (!PyString_Check(s)) {
	  PY_SET_ERROR(PyExc_ValueError, "list items must be strings");
	  goto error;
	}
	if (i == 0) sbuff = PyString_AS_STRING(s);
	scounts[i] = PyString_GET_SIZE(s);
	displs[i]     = PyString_AS_STRING(s) - sbuff;
      }}
    
    error:

      if (PyErr_Occurred())
	{ int i; for (i = 0; i < size; i++) scounts[i] = -1; }
      
    }

    /* scatter string size */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Scatter(scounts, 1, MPI_INT,
				    &rcount, 1, MPI_INT,
				    root, comm) );
    
    if (rcount == -1) {
      if (PyErr_Occurred()) RAISE;
      else PY_RAISE(PyExc_RuntimeError, "in root processor");
    }
    
    /* create python string */
    result = PyString_FromStringAndSize(NULL, rcount);
    if (!result) RAISE;
    
    /* scatter string buffer */
    rbuff = PyString_AS_STRING(result);
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Scatterv(sbuff, scounts, displs, PICKLED_STRING,
				     rbuff, rcount,          PICKLED_STRING,
				     root,comm) );
    
    /* deallocate memory */
    PyMem_Del(scounts); scounts = NULL;
    PyMem_Del(displs);     displs     = NULL;
    
    /* return result */
    return result;
  
  }
  
  EXCEPT {
    PyMem_Del(scounts);
    PyMem_Del(displs);
    Py_XDECREF(result);
    return NULL;
  }    
  
}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Allgather                                                        */
/*------------------------------------------------------------------*/
static 
PyObject* 
allgather_string(PyObject* self, PyObject* args) {
  
  PyObject* input  = NULL;
  PyObject* result = NULL;
  
  MPI_Comm comm;
  
  char* sbuff  = NULL;
  int   scount = 0;

  char* rbuff   = NULL;
  int*  rcounts = NULL;
  int*  displs  = NULL; 

  /* parse input arguments */
  if (!PyArg_ParseTuple(args, "OO&",
			&input, 
			PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
 
  TRY {

    /* comm size */
    int size;
    MPI_CHECK( MPI_Comm_size(comm,&size) );
    
    /* check input object */
    if (!PyString_Check(input))
      PY_RAISE(PyExc_TypeError, "expecting a string.");
    sbuff   = PyString_AS_STRING(input);
    scount = PyString_GET_SIZE(input);
    
    /* allocate arrays 'rcounts' and 'displs' */
    rcounts = PyMem_New(int, size);
    displs     = PyMem_New(int, size);
    if ( !rcounts || !displs )
      PY_RAISE(PyExc_MemoryError, "");
    
    /* gather string sizes */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Allgather(&scount, 1, MPI_INT,
				      rcounts, 1, MPI_INT,
				      comm) );
    
    /* create result list */
    result = PyList_New(size);
    if ( !result ) RAISE;

    /* fill result list */
    { int i; for (i = 0; i < size; i++) {
      /* create python string */
      PyObject* s = PyString_FromStringAndSize(NULL, rcounts[i]);
      if (!s) RAISE;
      /* set string in result list */
      PyList_SET_ITEM(result,i,s);
      /* calculate 'rbuff' pointer and 'displs' array */
      if (i == 0) rbuff = PyString_AS_STRING(s);
      displs[i] = PyString_AS_STRING(s) - rbuff;
    }}
    
    /* gather string buffers */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Allgatherv(sbuff, scount,          PICKLED_STRING,
				       rbuff, rcounts, displs, PICKLED_STRING,
				       comm) );
    /* deallocate memory */
    PyMem_Del(rcounts);
    PyMem_Del(displs);
    
    /* return result */
    return  result;
    
  }
  
  EXCEPT {
    PyMem_Del(rcounts);
    PyMem_Del(displs);
    Py_XDECREF(result);
    return NULL;
  }
  
}
/*------------------------------------------------------------------*/


/*------------------------------------------------------------------*/
/* Alltoall                                                         */
/*------------------------------------------------------------------*/
static 
PyObject* 
alltoall_string(PyObject* self, PyObject* args) {

  PyObject* input  = NULL;
  PyObject* result = NULL;
  
  MPI_Comm comm;
 
  char* sbuff   = NULL;
  int*  scounts = NULL;
  int*  sdispls = NULL;
  
  char* rbuff   = NULL;
  int*  rcounts = NULL;
  int*  rdispls = NULL; 

  /* parse input arguments */
  if (!PyArg_ParseTuple(args, "OO&",
			&input, 
			PyObj2C, &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {

    /* comm size   */
    int size;
    MPI_CHECK( MPI_Comm_size(comm,&size) );

    /* allocate arrays 'scounts' and 'sdispls' */
    scounts = PyMem_New(int, size);
    sdispls    = PyMem_New(int, size);
    if ( !scounts || !sdispls )
      PY_RAISE(PyExc_MemoryError, "");
    
    /* check input object */
    if ( !PyList_Check(input) )
      PY_RAISE(PyExc_TypeError, "expecting a list");
    /* check list length & comm size */
    if (size != PyList_GET_SIZE(input) )
      PY_RAISE(PyExc_ValueError, "invalid list length");
    
    /* calculate 'sbuff', 'scounts' and 'sdispls' */
    { int i; for (i = 0; i < size; i++) {
      PyObject* s = PyList_GET_ITEM(input, i);
      if ( !PyString_Check(s) )
	PY_RAISE(PyExc_ValueError, "list items must be strings");
      if (i == 0) sbuff = PyString_AS_STRING(s);
      scounts[i] = PyString_GET_SIZE(s);
      sdispls[i]    = PyString_AS_STRING(s) - sbuff;
    }}
    
    /* allocate arrays 'rcounts' and 'rdispls' */
    rcounts = PyMem_New(int, size);
    rdispls    = PyMem_New(int, size);
    if ( !rcounts || !rdispls )
      PY_RAISE(PyExc_MemoryError, "");

    /* alltoall string sizes */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Alltoall(scounts, 1, MPI_INT,
				     rcounts, 1, MPI_INT,
				     comm) );

    /* create result list */
    result = PyList_New(size);
    if (!result) RAISE;
    
    /* fill result list */
    { int i; for (i = 0; i < size; i++) {
      /* create python string */
      PyObject* s = PyString_FromStringAndSize(NULL, rcounts[i]);
      if (!s) PY_RAISE(PyExc_MemoryError, "");
      /* set string in result list */
      PyList_SET_ITEM(result,i,s);
      /* calculate 'rbuff' pointer and 'displs' array */
      if (i == 0) rbuff = PyString_AS_STRING(s);
      rdispls[i] = PyString_AS_STRING(s) - rbuff;
    }}
    
    /* alltoall string buffers */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Alltoallv(sbuff, scounts, sdispls, PICKLED_STRING,
				      rbuff, rcounts, rdispls, PICKLED_STRING,
				      comm) );

    /* deallocate memory */
    PyMem_Del(scounts);
    PyMem_Del(sdispls);
    PyMem_Del(rcounts);
    PyMem_Del(rdispls);
    
    /* return result */
    return  result;
    
  }

  EXCEPT {
    PyMem_Del(scounts);
    PyMem_Del(sdispls);
    PyMem_Del(rcounts);
    PyMem_Del(rdispls);
    Py_XDECREF(result);
    return NULL;
  }

}
/*------------------------------------------------------------------*/





/*------------------------------------------------------------------*/
/* Environmental Management                                         */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/

#ifdef __cplusplus
extern "C" {
#endif
extern void Py_GetArgcArgv(int* argc, char*** argv);
#ifdef __cplusplus
}
#endif

/* module initialization flag */
static int module_called_init = 0;
/* module finalization flag */
static int module_called_finalize = 0;

static
int
mpi_finalize(void) {
  int flag_i, flag_f;
  MPI_Initialized(&flag_i);
  if (!flag_i) {
    PyErr_SetString(PyExc_RuntimeError,
		    "MPI_Init() has not been called");
    return -1;
  }
  MPI_Finalized(&flag_f);
  if ( !flag_f && module_called_init ) {
    int ierr = MPI_Finalize();
    if ( ierr != MPI_SUCCESS ) {
      PyErr_Format(PyExc_RuntimeError,
		   "MPI_Finalize() failed [error code: %d]",ierr);
      return -1;
    }
    module_called_finalize = 1;
  }
  return -0;
}


static
void
mpi_atexit(void) {
  int flag, ierr;
  MPI_Initialized(&flag);
  if (!flag || !module_called_init) return;
  MPI_Finalized(&flag);
  if (flag || module_called_finalize) return;
  ierr = MPI_Finalize();
  module_called_finalize = 1;
  if (ierr != MPI_SUCCESS) {
    fflush(stderr);
    fprintf(stderr,"MPI_Finalize() failed [error code: %d]",ierr);
    fflush(stderr);
  }
}

static
int
mpi_init(void) {
  int flag_i, flag_f;
  MPI_Finalized(&flag_f);
  if (flag_f) {
    PyErr_SetString(PyExc_RuntimeError,
		    "MPI_Finalize() has been already called");
    return -1;
  }
  MPI_Initialized(&flag_i);
  if (!flag_i) {
    /* initialize MPI */
    int ierr;
    int argc; char** argv;
    Py_GetArgcArgv(&argc, &argv);
    /* This does not work with MPICH !!! */
    ierr = MPI_Init(&argc,&argv);
    if (ierr != MPI_SUCCESS) {
      PyErr_Format(PyExc_RuntimeError,
		   "MPI_Init() failed [error code: %d]",ierr);
      return -1;
    }
    if (Py_AtExit(mpi_atexit) < 0)
      PyErr_Warn(PyExc_RuntimeWarning,
		 "cannot register MPI_Finalize() with Py_AtExit()");
    module_called_init = 1; /* set init flag */
  }
  return 0;
}


/*------------------------------------------------------------------*/
static
PyObject*
init(PyObject* self, PyObject* args) {

  /* try to initialize MPI */
  if (mpi_init() < 0) 
    return NULL;
  else
    Py_RETURN_NONE;

}
/*------------------------------------------------------------------*/
static
PyObject*
initialized(PyObject* self, PyObject* args) {

  int flag;
  MPI_Initialized(&flag);
  return PyBool_FromLong((long)flag);
  
}
/*------------------------------------------------------------------*/

static
PyObject*
finalize(PyObject* self, PyObject* args) {

  /* try to finalize MPI */
  if (mpi_finalize() < 0) 
    return NULL;
  else
    Py_RETURN_NONE;

}
/*------------------------------------------------------------------*/
static
PyObject*
finalized(PyObject* self, PyObject* args) {
  
  int flag;
  MPI_Finalized(&flag);
  return PyBool_FromLong((long)flag);

}
/*------------------------------------------------------------------*/
static
PyObject*
comm_abort(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  int errorcode;
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&i",
			 PyObj2C, &comm,
			 &errorcode ) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);

  TRY {
    /* abort */
    MPI_CHECK_CALL_COMM(comm, 
			MPI_Abort(comm,errorcode) );
    Py_RETURN_NONE;
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static
PyObject*
get_version(PyObject* self, PyObject* args) {
  
  int version, subversion;
  MPI_Get_version(&version, &subversion);
  return Py_BuildValue("ii",version,subversion);

}
/*------------------------------------------------------------------*/
static
PyObject*
get_proc_name(PyObject* self, PyObject* args) {
  
  TRY {
    char name[MPI_MAX_PROCESSOR_NAME] = { '\0' };
    int  len = 0;
    MPI_CHECK_CALL( MPI_Get_processor_name(name, &len) );
    return PyString_FromStringAndSize(name, len);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject*
wtime(PyObject* self, PyObject* args) {

  return PyFloat_FromDouble( MPI_Wtime() );

}
/*------------------------------------------------------------------*/
static 
PyObject*
wtick(PyObject* self, PyObject* args) {
  
  return PyFloat_FromDouble( MPI_Wtick() );

}
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject* 
comm_get_errhandler(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&",
			 PyObj2C,  &comm) ) return NULL;

  MPI_CHECK_NULL(comm, MPI_COMM_NULL, MPI_ERR_COMM);
  
  TRY {
    /* get errhandler */
    MPI_Errhandler errhandler;
#if MPI_VERSION == 2
    MPI_CHECK( MPI_Comm_get_errhandler(comm, &errhandler) );
#else
    MPI_CHECK( MPI_Errhandler_get(comm, &errhandler) );
#endif
    /* return errhandler */
    return PyMPIErrhandler_FromErrhandler(errhandler);
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject* 
comm_set_errhandler(PyObject* self, PyObject* args) {

  MPI_Comm comm;
  MPI_Errhandler errhandler;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&O&",
			 PyObj2C, &comm,
			 PyObj2E, &errhandler) ) return NULL;

  MPI_CHECK_NULL(comm,       MPI_COMM_NULL,       MPI_ERR_COMM);
  MPI_CHECK_NULL(errhandler, MPI_ERRHANDLER_NULL, MPI_ERR_ARG);

  TRY {
    /* set errhandler */
#if MPI_VERSION == 2
    MPI_CHECK( MPI_Comm_set_errhandler(comm, errhandler) );
#else
    MPI_CHECK( MPI_Errhandler_set(comm, errhandler) );
#endif
    Py_RETURN_NONE;
  }
  
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject* 
errhandler_free(PyObject* self, PyObject* args) {

  MPI_Errhandler* errhandler;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O&", 
			 PyObj2EP, &errhandler) ) return NULL;
  
  MPI_CHECK_NULL(*errhandler, MPI_ERRHANDLER_NULL, MPI_ERR_ARG);

  TRY {
    /* free communicator */
    MPI_CHECK_CALL_FREE( MPI_Errhandler_free(errhandler) );
    Py_RETURN_NONE;
  }

  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/
static 
PyObject* 
error_string(PyObject* self, PyObject* args) {

  int errorcode;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "i", 
			 &errorcode) ) return NULL;
  
  TRY {
    char buf[MPI_MAX_ERROR_STRING];
    int  len;
    /* check errorcode range */
    if (errorcode<MPI_SUCCESS || 
	errorcode>=MPI_ERR_LASTCODE)
      PY_RAISE(PyExc_ValueError, 
	       "invalid error code, value out of range");
    /* get error string */
    MPI_CHECK_CALL( MPI_Error_string(errorcode, buf, &len) );
    /* return result */
#if defined(LAM_MPI)
    return Py_BuildValue("s#", buf+18, len-18);
#else
    return Py_BuildValue("s#", buf, len);
#endif
  }
  EXCEPT {
    return NULL;
  }
  
}
/*------------------------------------------------------------------*/
static 
PyObject* 
error_class(PyObject* self, PyObject* args) {

  int errorcode;

  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "i", 
			 &errorcode) ) return NULL;
  
  TRY {
    int errorclass;
    /* check errorcode range */
    if (errorcode<MPI_SUCCESS || 
	errorcode>=MPI_ERR_LASTCODE)
      PY_RAISE(PyExc_ValueError,
	       "invalid error code, value out of range");
    /* get error class */
    MPI_CHECK_CALL( MPI_Error_class(errorcode,&errorclass) );
    /* return result */
    return PyInt_FromLong((long)errorclass);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/




/*------------------------------------------------------------------*/
/* Misc                                                             */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/
static 
PyObject* 
_set_exception(PyObject* self, PyObject* args) {

  PyObject* exc;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "O",
			 &exc) ) return NULL;
  
  /* set exception class */
  if (exc != PyExc_MPIError) {
    Py_XDECREF(PyExc_MPIError);
    PyExc_MPIError = exc;
    Py_INCREF(PyExc_MPIError);
  }
  Py_RETURN_NONE;
  
}
/*------------------------------------------------------------------*/
static 
PyObject* 
_raise(PyObject* self, PyObject* args) {

  int errorcode;
  
  /* parse input arguments */
  if ( !PyArg_ParseTuple(args, "i", 
			 &errorcode) ) return NULL;

  TRY {
    /* check errorcode range */
    if (errorcode<MPI_SUCCESS || 
	errorcode>=MPI_ERR_LASTCODE)
      PY_RAISE(PyExc_ValueError,
	       "invalid error code, value out of range");
    MPI_CHECK( errorcode );
    Py_RETURN_NONE;
  }
  EXCEPT {
    return NULL;
  }
  
}
/*------------------------------------------------------------------*/
static 
PyObject* 
distribute(PyObject* self, PyObject* args) {
  
  int N,B,i;
  
  /* parse input arguments */
  if (!PyArg_ParseTuple(args, "iii",
			&N,&B,&i) ) return NULL;
  
  TRY {

    int low,high;
    
    if ( N < 0 )
      PY_RAISE(PyExc_ValueError,"argument 1 less than zero");
    if ( B < 1 )
      PY_RAISE(PyExc_ValueError,"argument 2 less than one");
    if ( i < 0 || i >= B )
      PY_RAISE(PyExc_ValueError,"argument 3 out of range");

    {    
      int C = N / B; /* integer division */
      int R = N % B; /* remainder */
      if (i < R) {
	low  = i*C + i;
	high = low + C + 1;
      }
      else {
	low  = i*C + R;
	high = low + C;
      }
    }
    return Py_BuildValue("ii", low, high);
  }
  EXCEPT {
    return NULL;
  }

}
/*------------------------------------------------------------------*/






/*------------------------------------------------------------------*/
/* Python Module                                                    */
/*------------------------------------------------------------------*/

/*------------------------------------------------------------------*/

PyDoc_STRVAR(mpi_doc, "MPI extension module");

/*------------------------------------------------------------------*/

static struct PyMethodDef mpi_methods[] = {

  /* Comm - Intracomm */
  {"comm_size",         comm_size,         METH_VARARGS},
  {"comm_rank",         comm_rank,         METH_VARARGS},
  {"comm_compare",      comm_compare,      METH_VARARGS},
  {"comm_dup",          comm_dup,          METH_VARARGS},
  {"comm_create",       comm_create,       METH_VARARGS},
  {"comm_split",        comm_split,        METH_VARARGS},
  {"comm_free",         comm_free,         METH_VARARGS},

  /* Group */
  {"group_size",        group_size,        METH_VARARGS},
  {"group_rank",        group_rank,        METH_VARARGS},
  {"group_transl_rank", group_transl_rank, METH_VARARGS},
  {"group_compare",     group_compare,     METH_VARARGS},
  {"comm_group",        comm_group,        METH_VARARGS},
  {"group_union",       group_union,       METH_VARARGS},
  {"group_intersection",group_intersection,METH_VARARGS},
  {"group_difference",  group_difference,  METH_VARARGS},
  {"group_incl",        group_incl,        METH_VARARGS},
  {"group_excl",        group_excl,        METH_VARARGS},
  {"group_range_incl",  group_range_incl,  METH_VARARGS},
  {"group_range_excl",  group_range_excl,  METH_VARARGS},
  {"group_free",        group_free,        METH_VARARGS},

  /* Status */
  {"get_count",         status_get_count,  METH_VARARGS},
  {"get_elements",      status_get_elems,  METH_VARARGS},
  {"test_cancelled",    test_cancelled,    METH_VARARGS},


  /* Request */
  {"wait",              request_wait,      METH_VARARGS},
  {"test",              request_test,      METH_VARARGS},
  {"request_free",      request_free,      METH_VARARGS},
  {"cancel",            request_cancel,    METH_VARARGS},


  /* Datatype */
  {"type_get_extent",   type_get_extent,   METH_VARARGS},
  {"type_extent",       type_extent,       METH_VARARGS},
  {"type_lb",           type_lb,           METH_VARARGS},
  {"type_ub",           type_ub,           METH_VARARGS},
  {"type_size",         type_size,         METH_VARARGS},
  {"type_contiguous",   type_contiguous,   METH_VARARGS},
  {"type_vector",       type_vector,       METH_VARARGS},
  {"type_commit",       type_commit,       METH_VARARGS},
  {"type_free",         type_free,         METH_VARARGS},

  /* Point-to-Point Communications */
  {"send_string",       send_string,       METH_VARARGS},
  {"recv_string",       recv_string,       METH_VARARGS},
  {"sendrecv_string",   sendrecv_string,   METH_VARARGS},
  {"comm_probe",        comm_probe,        METH_VARARGS},

  /* Collective Communications */
  {"barrier",           barrier,           METH_VARARGS},
		        		   
  {"bcast_string",      bcast_string,      METH_VARARGS},
  {"scatter_string",    scatter_string,    METH_VARARGS},
  {"gather_string",     gather_string,     METH_VARARGS},
  {"allgather_string",  allgather_string,  METH_VARARGS},
  {"alltoall_string",   alltoall_string,   METH_VARARGS},

  /* Intercomm */
  {"comm_test_inter",   comm_test_inter,   METH_VARARGS},
  {"comm_remote_size",  comm_remote_size,  METH_VARARGS},
  {"comm_remote_group", comm_remote_group, METH_VARARGS},
  {"intercomm_create",  intercomm_create,  METH_VARARGS},
  {"intercomm_merge",   intercomm_merge,   METH_VARARGS},

  /* Topologies */
  {"topo_test",         topo_test,         METH_VARARGS},
  {"cart_create",       cart_create,       METH_VARARGS},
  {"dims_create",       dims_create,       METH_VARARGS},
  {"cartdim_get",       cartdim_get,       METH_VARARGS},
  {"cart_get",          cart_get,          METH_VARARGS},
  {"cart_rank",         cart_rank,         METH_VARARGS},
  {"cart_coords",       cart_coords,       METH_VARARGS},
  {"cart_shift",        cart_shift,        METH_VARARGS},
  {"cart_sub",          cart_sub,          METH_VARARGS},
  {"cart_map",          cart_map,          METH_VARARGS},

  {"graph_create",      graph_create,      METH_VARARGS},
  {"graphdims_get",     graphdims_get,     METH_VARARGS},
  {"graph_get",         graph_get,         METH_VARARGS},
  {"graph_neigh_count", graph_neigh_count, METH_VARARGS},
  {"graph_neigh",       graph_neigh,       METH_VARARGS},
  {"graph_map",         graph_map,         METH_VARARGS},
		        		   
  /* Environmental Management */
  {"init",              init,              METH_NOARGS},
  {"initialized",       initialized,       METH_NOARGS},
  {"finalize",          finalize,          METH_NOARGS},
  {"finalized",         finalized,         METH_NOARGS},
  {"comm_abort",        comm_abort,        METH_VARARGS},
  
  {"get_version",       get_version,       METH_NOARGS},
  {"get_processor_name",get_proc_name,     METH_NOARGS},
  
  {"wtime",             wtime,             METH_NOARGS},
  {"wtick",             wtick,             METH_NOARGS},
  
  {"comm_get_errhandler", comm_get_errhandler, METH_VARARGS},
  {"comm_set_errhandler", comm_set_errhandler, METH_VARARGS},
  {"errhandler_free",     errhandler_free,     METH_VARARGS},

  {"error_class",       error_class,       METH_VARARGS},
  {"error_string",      error_string,      METH_VARARGS},

  /* Misc */
  {"_set_exception",    _set_exception,    METH_VARARGS},
  {"_raise",            _raise,            METH_VARARGS},
  {"distribute",        distribute,        METH_VARARGS},

  {NULL, NULL} /* Sentinel */

};

/*------------------------------------------------------------------*/

#define MPI4PY_ADD_TYPE(m,name,type)                 \
  { Py_INCREF(PyMPI##type##_Type);                   \
    if (PyModule_AddObject(m,name,                   \
       (PyObject*) PyMPI##type##_Type) < 0) return; }

#define MPI4PY_ADD_CONSTANT(m,name,value) \
  { if (PyModule_AddIntConstant(m,name,value) < 0) return; }

#define MPI4PY_ADD_OBJ(m,name,type,value)           \
  { PyObject* ob = PyMPI##type##_From##type(value); \
    if (PyModule_AddObject(m,name,ob) < 0) return; }

/*------------------------------------------------------------------*/

PyMODINIT_FUNC
init_mpi(void) {

  PyObject* m;


  /* Module Initialization */
  /* --------------------- */
  m = Py_InitModule4("_mpi", mpi_methods, mpi_doc,
		     NULL, PYTHON_API_VERSION);
  if (m == NULL) return;

  /* MPI Initialization */
  /* ------------------ */
  if (import_libmpi() < 0) return;
  if (mpi_init()      < 0) return;



  /* Constants */
  /* --------- */  
  {
    MPI4PY_ADD_CONSTANT(m, "ANY_TAG"    , MPI_ANY_TAG    );
    MPI4PY_ADD_CONSTANT(m, "ANY_SOURCE" , MPI_ANY_SOURCE );
    MPI4PY_ADD_CONSTANT(m, "PROC_NULL"  , MPI_PROC_NULL  );
    MPI4PY_ADD_CONSTANT(m, "UNDEFINED"  , MPI_UNDEFINED  );
  }
  

  /* Group */
  /* ----- */
  {
    /* type */
    MPI4PY_ADD_TYPE(m, "Group", Group);
    /* null handle */
    MPI4PY_ADD_OBJ(m, "GROUP_NULL" , Group, MPI_GROUP_NULL  );
    /* predefined */		     	    
    MPI4PY_ADD_OBJ(m, "GROUP_EMPTY", Group, MPI_GROUP_EMPTY );
  }
  

  /* Communicator */
  /* ------------ */
  {
    /* type */
    MPI4PY_ADD_TYPE(m, "Comm", Comm);
    /* null handle */
    MPI4PY_ADD_OBJ(m, "COMM_NULL" , Comm, MPI_COMM_NULL  );
    /* predefined */		   	   
    MPI4PY_ADD_OBJ(m, "COMM_WORLD", Comm, MPI_COMM_WORLD ); 
    MPI4PY_ADD_OBJ(m, "COMM_SELF" , Comm, MPI_COMM_SELF  );  
    /* topologies */
    MPI4PY_ADD_CONSTANT(m, "CART"  , MPI_CART  );
    MPI4PY_ADD_CONSTANT(m, "GRAPH" , MPI_GRAPH );
    /* comparisons */
    MPI4PY_ADD_CONSTANT(m, "IDENT"     , MPI_IDENT     );
    MPI4PY_ADD_CONSTANT(m, "CONGRUENT" , MPI_CONGRUENT );
    MPI4PY_ADD_CONSTANT(m, "SIMILAR"   , MPI_SIMILAR   );
    MPI4PY_ADD_CONSTANT(m, "UNEQUAL"   , MPI_UNEQUAL   );
  }


  /* Status */
  /* ------ */
  {
    /* type */
    MPI4PY_ADD_TYPE(m, "Status", Status);
  }


  /* Request */
  /* ------- */
  {
    /* type */
    MPI4PY_ADD_TYPE(m, "Request", Request);
    /* null handle */
    MPI4PY_ADD_OBJ(m, "REQUEST_NULL", Request, MPI_REQUEST_NULL );
  }


  /* Datatype */
  /* ------ */
  {
    /* type */
    MPI4PY_ADD_TYPE(m, "Datatype", Datatype);
    /* null handle */
    MPI4PY_ADD_OBJ(m, "DATATYPE_NULL"  , Datatype,  MPI_DATATYPE_NULL   );
    /* predefined */  				  
    MPI4PY_ADD_OBJ(m, "CHAR"           , Datatype,  MPI_CHAR            );
    MPI4PY_ADD_OBJ(m, "SHORT"          , Datatype,  MPI_SHORT           );
    MPI4PY_ADD_OBJ(m, "INT"            , Datatype,  MPI_INT             );
    MPI4PY_ADD_OBJ(m, "LONG"           , Datatype,  MPI_LONG            );
    MPI4PY_ADD_OBJ(m, "UNSIGNED_CHAR"  , Datatype,  MPI_UNSIGNED_CHAR   );
    MPI4PY_ADD_OBJ(m, "UNSIGNED_SHORT" , Datatype,  MPI_UNSIGNED_SHORT  );
    MPI4PY_ADD_OBJ(m, "UNSIGNED"       , Datatype,  MPI_UNSIGNED        );
    MPI4PY_ADD_OBJ(m, "UNSIGNED_LONG"  , Datatype,  MPI_UNSIGNED_LONG   );
    MPI4PY_ADD_OBJ(m, "FLOAT"          , Datatype,  MPI_FLOAT           );
    MPI4PY_ADD_OBJ(m, "DOUBLE"         , Datatype,  MPI_DOUBLE          );
    MPI4PY_ADD_OBJ(m, "LONG_DOUBLE"    , Datatype,  MPI_LONG_DOUBLE     );
    MPI4PY_ADD_OBJ(m, "BYTE"           , Datatype,  MPI_BYTE            );
    MPI4PY_ADD_OBJ(m, "PACKED"         , Datatype,  MPI_PACKED          );
    MPI4PY_ADD_OBJ(m, "FLOAT_INT"      , Datatype,  MPI_FLOAT_INT       );
    MPI4PY_ADD_OBJ(m, "DOUBLE_INT"     , Datatype,  MPI_DOUBLE_INT      );
    MPI4PY_ADD_OBJ(m, "LONG_INT"       , Datatype,  MPI_LONG_INT        );
    MPI4PY_ADD_OBJ(m, "TWOINT"         , Datatype,  MPI_2INT	        );
    MPI4PY_ADD_OBJ(m, "SHORT_INT"      , Datatype,  MPI_SHORT_INT       );
    MPI4PY_ADD_OBJ(m, "LONG_DOUBLE_INT", Datatype,  MPI_LONG_DOUBLE_INT );
    MPI4PY_ADD_OBJ(m, "UB"             , Datatype,  MPI_UB	        ); 
    MPI4PY_ADD_OBJ(m, "LB"             , Datatype,  MPI_LB              );
  }


  /* Op */
  /* -- */
  {
    /* type */
    MPI4PY_ADD_TYPE(m, "Op", Op);
    /* null handle */
    MPI4PY_ADD_OBJ(m, "OP_NULL", Op, MPI_OP_NULL );
    /* predefined */
    MPI4PY_ADD_OBJ(m, "MAX"    , Op, MPI_MAX     );
    MPI4PY_ADD_OBJ(m, "MIN"    , Op, MPI_MIN     );
    MPI4PY_ADD_OBJ(m, "SUM"    , Op, MPI_SUM     );
    MPI4PY_ADD_OBJ(m, "PROD"   , Op, MPI_PROD    );
    MPI4PY_ADD_OBJ(m, "LAND"   , Op, MPI_LAND    );
    MPI4PY_ADD_OBJ(m, "BAND"   , Op, MPI_BAND    );
    MPI4PY_ADD_OBJ(m, "LOR"    , Op, MPI_LOR     );
    MPI4PY_ADD_OBJ(m, "BOR"    , Op, MPI_BOR     );
    MPI4PY_ADD_OBJ(m, "LXOR"   , Op, MPI_LXOR    );
    MPI4PY_ADD_OBJ(m, "BXOR"   , Op, MPI_BXOR    );
    MPI4PY_ADD_OBJ(m, "MAXLOC" , Op, MPI_MAXLOC  );
    MPI4PY_ADD_OBJ(m, "MINLOC" , Op, MPI_MINLOC  );
  }
  

  /* Environmental inquires */
  /* ---------------------- */
  {
    int* tagub  = NULL;
    int* host   = NULL;
    int* io     = NULL;
    int* wtglb  = NULL;
    
    int flag;
    MPI_Attr_get(MPI_COMM_WORLD, MPI_TAG_UB          , &tagub  , &flag);
    MPI_Attr_get(MPI_COMM_WORLD, MPI_HOST            , &host   , &flag);
    MPI_Attr_get(MPI_COMM_WORLD, MPI_IO              , &io     , &flag);
    MPI_Attr_get(MPI_COMM_WORLD, MPI_WTIME_IS_GLOBAL , &wtglb  , &flag);
    
    MPI4PY_ADD_CONSTANT(m, "TAG_UB"          , *tagub );
    MPI4PY_ADD_CONSTANT(m, "HOST"            , *host  );
    MPI4PY_ADD_CONSTANT(m, "IO"              , *io    );
    MPI4PY_ADD_CONSTANT(m, "WTIME_IS_GLOBAL" , *wtglb );
  }
  

  /* Default Exception */
  /* ----------------- */
  {
    PyExc_MPIError = PyExc_RuntimeError;
    Py_INCREF(PyExc_MPIError);
  }


  /* Errhandler */
  /* ---------- */
  {
    /* type */
    MPI4PY_ADD_TYPE(m, "Errhandler", Errhandler);
    /* null handle */
    MPI4PY_ADD_OBJ(m, "ERRHANDLER_NULL" , Errhandler, MPI_ERRHANDLER_NULL  );
    /* predefined */			  		  
    MPI4PY_ADD_OBJ(m, "ERRORS_ARE_FATAL", Errhandler, MPI_ERRORS_ARE_FATAL );
    MPI4PY_ADD_OBJ(m, "ERRORS_RETURN"   , Errhandler, MPI_ERRORS_RETURN    );
  }

  
  /* Error Codes */
  /* ----------- */
  {
    MPI4PY_ADD_CONSTANT(m, "SUCCESS"       , MPI_SUCCESS       );
    MPI4PY_ADD_CONSTANT(m, "ERR_ARG"       , MPI_ERR_ARG       );
    MPI4PY_ADD_CONSTANT(m, "ERR_BUFFER"    , MPI_ERR_BUFFER    );
    MPI4PY_ADD_CONSTANT(m, "ERR_COMM"      , MPI_ERR_COMM      );
    MPI4PY_ADD_CONSTANT(m, "ERR_COUNT"     , MPI_ERR_COUNT     );
    MPI4PY_ADD_CONSTANT(m, "ERR_DIMS"      , MPI_ERR_DIMS      );
    MPI4PY_ADD_CONSTANT(m, "ERR_GROUP"     , MPI_ERR_GROUP     );
    MPI4PY_ADD_CONSTANT(m, "ERR_IN_STATUS" , MPI_ERR_IN_STATUS );
    MPI4PY_ADD_CONSTANT(m, "ERR_INTERN"    , MPI_ERR_INTERN    );
    MPI4PY_ADD_CONSTANT(m, "ERR_LASTCODE"  , MPI_ERR_LASTCODE  );
    MPI4PY_ADD_CONSTANT(m, "ERR_OP"        , MPI_ERR_OP        );
    MPI4PY_ADD_CONSTANT(m, "ERR_OTHER"     , MPI_ERR_OTHER     );
    MPI4PY_ADD_CONSTANT(m, "ERR_PENDING"   , MPI_ERR_PENDING   );
    MPI4PY_ADD_CONSTANT(m, "ERR_RANK"      , MPI_ERR_RANK      );
    MPI4PY_ADD_CONSTANT(m, "ERR_REQUEST"   , MPI_ERR_REQUEST   );
    MPI4PY_ADD_CONSTANT(m, "ERR_ROOT"      , MPI_ERR_ROOT      );
    MPI4PY_ADD_CONSTANT(m, "ERR_TAG"       , MPI_ERR_TAG       );
    MPI4PY_ADD_CONSTANT(m, "ERR_TOPOLOGY"  , MPI_ERR_TOPOLOGY  );
    MPI4PY_ADD_CONSTANT(m, "ERR_TRUNCATE"  , MPI_ERR_TRUNCATE  );
    MPI4PY_ADD_CONSTANT(m, "ERR_TYPE"      , MPI_ERR_TYPE      );
    MPI4PY_ADD_CONSTANT(m, "ERR_UNKNOWN"   , MPI_ERR_UNKNOWN   );
  }
  

  /* Convenience Constants */
  /* --------------------- */
  {
    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    MPI4PY_ADD_CONSTANT(m, "rank" , rank );
    MPI4PY_ADD_CONSTANT(m, "size" , size );
  }


}

/*------------------------------------------------------------------*/
