# Author:    Lisandro Dalcin
# Contact:   dalcinl@users.sourceforge.net
# Revision:  $Revision: 1.5 $
# Date:      $Date: 2006/08/11 00:43:07 $
# Copyright: This module has been placed in the public domain.

# --------------------------------------------------------

# Environmental variables to look for configuration
MPICC_ENV  = ['MPI4PY_CC',  'MPICC']
MPICXX_ENV = ['MPI4PY_CXX', 'MPICXX']
MPICFG_ENV = ['MPI4PY_CFG', 'MPI_CONFIG', 'MPI_CFG']

# Default values to use for configuration
MPICC  = ['mpicc']
MPICXX = ['mpicxx', 'mpiCC', 'mpic++']
MPICFG = ('mpi', 'mpi.cfg')

# --------------------------------------------------------

import os
from distutils import sysconfig
from distutils.spawn import find_executable
from distutils import log


def customize_compiler(compiler, environ=None):
    """Do any platform-specific customization of a CCompiler instance.

    Mainly needed on Unix, so we can plug in the information that
    varies across Unices and is stored in Python's Makefile.
    """
    if environ is None:
        environ = os.environ
    if compiler.compiler_type == 'unix':
        (cc, cxx, opt, basecflags, ccshared, ldshared, so_ext) = \
            sysconfig.get_config_vars('CC', 'CXX', 'OPT', 'BASECFLAGS', 'CCSHARED', 'LDSHARED', 'SO')

        if environ.has_key('CC'):
            cc = environ['CC']
        if environ.has_key('CXX'):
            cxx = environ['CXX']
        if environ.has_key('LDSHARED'):
            ldshared = environ['LDSHARED']
        if environ.has_key('CPP'):
            cpp = environ['CPP']
        else:
            cpp = cc + " -E"           # not always
        if environ.has_key('LDFLAGS'):
            ldshared = ldshared + ' ' + environ['LDFLAGS']
        if basecflags:
            opt = basecflags + ' ' + opt
        if environ.has_key('CFLAGS'):
            opt = opt + ' ' + environ['CFLAGS']
            ldshared = ldshared + ' ' + environ['CFLAGS']
        if environ.has_key('CPPFLAGS'):
            cpp = cpp + ' ' + environ['CPPFLAGS']
            opt = opt + ' ' + environ['CPPFLAGS']
            ldshared = ldshared + ' ' + environ['CPPFLAGS']

        cc_cmd = cc + ' ' + opt
        compiler.set_executables(
            preprocessor=cpp,
            compiler=cc_cmd,
            compiler_so=cc_cmd + ' ' + ccshared,
            compiler_cxx=cxx,
            linker_so=ldshared,
            linker_exe=cc)

        compiler.shared_lib_extension = so_ext

def customize_mpi_environ(mpicc, mpicxx=None, environ=None):
    if environ is None:
        environ = dict(os.environ)
    if mpicc: # C compiler
        environ['CC'] = mpicc
    if mpicxx: # C++ compiler
        environ['CXX'] = mpicxx
    mpild = mpicc or mpicxx
    if mpild: # linker for shared
        ldshared = sysconfig.get_config_var('LDSHARED')
        if not ldshared:
            environ['LDSHARED'] = mpild
        else:
            ldshared = ldshared.split(' ', 1)
            if len(ldshared) == 2:
                environ['LDSHARED'] = mpild + ' ' + ldshared[1]
    return environ

def _find_mpi_compiler(envvars, executables, path=None):
    # search in environment
    if envvars:
        if isinstance(envvars, str):
            envvars = (envvars,)
        for var in envvars:
            cmd = os.environ.get(var)
            if cmd is not None: return cmd
    # search in path
    if executables:
        if isinstance(executables, str):
            executables = (executables,)
        for exe in executables:
            cmd = find_executable(exe, path)
            if cmd is not None: return cmd
    # nothing found
    return None


# --------------------------------------------------------

from ConfigParser import ConfigParser
from ConfigParser import Error as ConfigParserError

def _config_parser(section, filenames, raw=False, vars=None):
    """
    Returns a dictionary of options obtained by parsing configuration
    files.
    """
    parser = ConfigParser()
    try:
        parser.read(filenames.split(','))
    except ConfigParserError:
        log.error("error: parsing configuration file/s '%s'", filenames)
        return None
    if not parser.has_section(section):
        log.error("error: section '%s' not found "
                  "in configuration file/s '%s'", section, filenames)
        return None
    config_info = {}
    for k, v in parser.items(section, raw, vars):
        config_info[k] = v.replace('\\',' ').split()
    if 'define_macros' in config_info:
        macros = []
        for m in config_info['define_macros'] :
            try: # "-DFOO=blah"
                idx = m.index("=")
                macro = (m[:idx], m[idx+1:] or None)
            except ValueError: # bare "-DFOO"
                macro = (m, None)
            macros.append(macro)
        config_info['define_macros'] = macros
    return config_info


def _find_mpi_config(section, filenames,
                     envvars=None, defaults=None):
    if not section and not filenames and envvars:
        # look in environment
        if isinstance(envvars, str):
            envvars = (envvars,)
        for var in envvars:
            if var in os.environ:
                section, filenames = os.environ[var], None
                break
    if section and ',' in section:
        section, filenames = section.split(',', 1)
    if defaults:
        if not section:
            section = defaults[0]
        if not filenames:
            fname = defaults[1]
            if os.path.exists(fname):
                filenames = fname
    # parse configuraration
    if section and filenames:
        config_info = _config_parser(section, filenames)
        return section, filenames, config_info
    else:
        return section, filenames, None


def _configure(extension, confdict):
    if confdict is None: return
    for key, value in confdict.iteritems():
        if hasattr(extension, key):
            item = getattr(extension, key)
            if type(item) is list:
                if type(value) is list:
                    for v in value:
                        if v not in item:
                            item.append(v)
                else:
                    if value not in item:
                        item.append(value)
            else:
                setattr(extension, key, value)

# --------------------------------------------------------

from distutils.command import config as cmd_config
from distutils.command import build as cmd_build
from distutils.command import build_ext as cmd_build_ext
from distutils.command import sdist as cmd_sdist
from distutils.errors import DistutilsPlatformError


_ConfigTest = """\
int main(int argc, char **argv) {
  MPI_Init(&argc,&argv);
  MPI_Finalize();
  return 0;
}
"""

mpi_cmd_opts = [

    ('mpicc=',   None,
     "MPI C compiler command, "
     "overrides environmental variables 'MPICC' "
     "(defaults to 'mpicc' if available)"),

    ('mpicxx=',  None,
     "MPI C++ compiler command, "
     "overrides environmental variables 'MPICXX' "
     "(defaults to 'mpicxx', 'mpiCC', or 'mpic++' if any is available)"),

    ('mpi-cfg=', None,
     "specify a configuration file to look for MPI includes/libraries "
     "(defaults to 'mpi.cfg')"),

    ('mpi=',     None,
     "specify a configuration section, "
     "and an optional comma-separated list of configuration files "
     "(e.g. --mpi=section,file1,file2,file3),"
     "to look for MPI includes/libraries, "
     "overrides environmental variables 'MPICFG' "
     "(defaults to section 'mpi' in configuration file 'mpi.cfg')"),
    
    ]


class config(cmd_config.config):

    user_options = cmd_config.config.user_options + mpi_cmd_opts

    def initialize_options(self):
        cmd_config.config.initialize_options(self)
        self.mpicc = None
        self.mpicxx = None
        self.mpi = None
        self.mpi_cfg = None

    def find_mpi_compiler(self, envvars, executables, path=None):
        return _find_mpi_compiler(envvars, executables, path)

    def run(self):
        # test MPI C compiler
        mpicc = self.mpicc
        if mpicc is None:
            mpicc = self.find_mpi_compiler(MPICC_ENV, MPICC)
        if mpicc:
            log.info("MPI C compiler:    %s", mpicc  or 'not found')
            self.compiler = None
            self._check_compiler()
            environ = customize_mpi_environ(mpicc, None)
            customize_compiler(self.compiler, environ)
            self.try_link(_ConfigTest, headers=['mpi.h'], lang='c')
        # test MPI C++ compiler
        mpicxx = self.mpicxx
        if mpicxx is None:
            mpicxx = self.find_mpi_compiler(MPICXX_ENV, MPICXX)
        if mpicxx:
            log.info("MPI C++ compiler:  %s", mpicxx or 'not found')
            self.compiler = None
            self._check_compiler()
            environ = customize_mpi_environ(None, mpicxx)
            customize_compiler(self.compiler, environ)
            if self.compiler.compiler_type in ('unix', 'cygwin', 'mingw32'):
                self.compiler.compiler_so[0] = mpicxx
                self.compiler.linker_exe[0] = mpicxx
            self.try_link(_ConfigTest, headers=['mpi.h'], lang='c++')
        # test configuration in specified section and file
        if self.mpi or self.mpi_cfg:
            sct, fn, cfg = _find_mpi_config(self.mpi, self.mpi_cfg,
                                            MPICFG_ENV, MPICFG)
            log.info("MPI configuration: "
                     "section '%s' from file/s '%s'", sct, fn)
            _configure(self, cfg)
            self.compiler = None
            self._check_compiler()
            self.try_link(_ConfigTest, headers=['mpi.h'], lang='c')


class build(cmd_build.build):

    user_options = cmd_build.build.user_options + mpi_cmd_opts

    def initialize_options(self):
        cmd_build.build.initialize_options(self)
        self.mpicc = None
        self.mpicxx = None
        self.mpi = None
        self.mpi_cfg = None

    def finalize_options(self):
        cmd_build.build.finalize_options(self)
        config_cmd = self.get_finalized_command('config')
        if isinstance(config_cmd,  config):
            self.set_undefined_options('config',
                                       ('mpicc',   'mpicc'),
                                       ('mpicxx',  'mpicxx'),
                                       ('mpi',     'mpi'),
                                       ('mpi_cfg', 'mpi_cfg'))
        

class build_ext(cmd_build_ext.build_ext):

    user_options = cmd_build_ext.build_ext.user_options + mpi_cmd_opts

    def initialize_options(self):
        cmd_build_ext.build_ext.initialize_options(self)
        self.mpicc = None
        self.mpicxx = None
        self.mpi = None
        self.mpi_cfg = None

    def finalize_options(self):
        cmd_build_ext.build_ext.finalize_options(self)
        build_cmd = self.get_finalized_command('build')
        if isinstance(build_cmd,  build):
            self.set_undefined_options('build',
                                       ('mpicc',   'mpicc'),
                                       ('mpicxx',  'mpicxx'),
                                       ('mpi',     'mpi'),
                                       ('mpi_cfg', 'mpi_cfg'))

    def build_extensions(self):
        # First, sanity-check the 'extensions' list
        self.check_extensions_list(self.extensions)
        # configure
        self.configure_compiler()
        self.configure_extensions()
        # build
        for ext in self.extensions:
            self.build_extension(ext)

    def configure_compiler(self):
        mpicc, mpicxx = self.mpicc, self.mpicxx
        if mpicc is None:
            mpicc = self.find_mpi_compiler(MPICC_ENV, MPICC)
        if mpicxx is None:
            mpicxx = self.find_mpi_compiler(MPICXX_ENV, MPICXX)
        log.info("MPI C compiler:    %s", mpicc  or 'not found')
        log.info("MPI C++ compiler:  %s", mpicxx or 'not found')
        environ = customize_mpi_environ(mpicc, mpicxx)
        customize_compiler(self.compiler, environ)

    def find_mpi_compiler(self, envvars, executables, path=None):
        return _find_mpi_compiler(envvars, executables, path)

    def configure_extensions(self):
        # parse configuration file
        section, filenames = self.mpi, self.mpi_cfg
        config_info = self.find_mpi_config(section, filenames,
                                           MPICFG_ENV, MPICFG)
        # configure extensions
        if config_info:
            for ext in self.extensions:
                self.configure_extension(ext, config_info)

    def find_mpi_config(self, section, filenames,
                        envvars=None, defaults=None):
        # parse configuration file
        sect, fnames, cfg_info = _find_mpi_config(section, filenames,
                                                  envvars, defaults)
        if cfg_info:
            log.info("MPI configuration:  "
                     "from section '%s' in file/s '%s'", sect, fnames)
        return cfg_info
        
    def configure_extension(self, extension, config_info):
        _configure(extension, config_info)

    def find_swig(self):
        try:
            return cmd_build_ext.build_ext.find_swig(self)
        except DistutilsPlatformError:
            return 'swig'

    def spawn(self, cmd, search_path=1, level=1):
        if 'swig' in cmd[0]:
            swig = find_executable(cmd[0])
            if swig is not None:
                cmd.insert(2, '-O')
                cmd.insert(2, '-noproxy')
            else:
                log.info("swigging of %s skipped (SWIG not found)", cmd[-1])
                return
        cmd_build_ext.build_ext.spawn(self, cmd, search_path, level)
    

class sdist(cmd_sdist.sdist):
    
    def build_sources(self):
        build_ext_cmd = self.get_finalized_command('build_ext')
        if not isinstance(build_ext_cmd, build_ext): return
        build_ext_cmd.check_extensions_list(build_ext_cmd.extensions)
        for ext in build_ext_cmd.extensions:
            if sysconfig.get_python_version() < '2.4':
                build_ext_cmd.swig_sources(list(ext.sources))
            else:
                build_ext_cmd.swig_sources(list(ext.sources), ext)
    def run(self):
        self.build_sources()
        cmd_sdist.sdist.run(self)

